//	--------------------------------------------------------------------------
//
//	BITMAPS.CPP
//
//	--------------------------------------------------------------------------
#include	"bitmaps.hpp"
#include	"pcx.hpp"

BitmapResourceFile::BitmapResourceFile(DirString* pathname,uint options,long num_resources)
					: ResourceFile(pathname,options,num_resources,resource_BITMAP_RESOURCE)
	{
	}

BitmapResourceFile::BitmapResourceFile(	const char* pathname,uint options,long num_resources)
					: ResourceFile(	pathname,options,num_resources,resource_BITMAP_RESOURCE)
	{
	}

BitmapResourceFile::~BitmapResourceFile()
	{
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	SizeOfDecompressedResource
//	DESC:	use to malloc the size of image in a resource
//
//	-------------------------------------------------------------------------
long BitmapResourceFile::SizeOfDecompressedResource (long resource_num)
	{
	long		fileError;
	// read the resource index
	if ((fileError = ReadResourceIndex( resource_num )) != resource_NoError) return fileError;

	// if the size is zero, there is no said resource
	if (index.size==0) return fileError = resource_ResourceNotFound;

	// seek to the start of the resource
	if ((fileError = Seek( index.location )) != resource_NoError) return fileError;

	long size = 0;

	// read in the PCX_resourceHeader
	PCX_resourceHeader hdr;
	DOSFile::Read((char*) &hdr, sizeof(PCX_resourceHeader));

	// read the resource decompressed size and return it
	size = hdr.sizeofDecompressed;
	return size;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	DecompressResource
//	DESC:	obtains an image from a resource by calling an appropriate
//			decompress routine
//
//	-------------------------------------------------------------------------
int BitmapResourceFile::DecompressResource(char* dest, char* srce)
	{
	// call appropriate decompression routine
	PCX_ResourceToImage((grImage*)dest, (PCX_resource*)srce);
	return true;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		ReadResource
//	DESC:       reads a resource of 'size' bytes, decompresses it and
//				places them in an image 'buffer'
//
//	-------------------------------------------------------------------------
int BitmapResourceFile::ReadResource(long resource_num,char* buffer)
	{
	// skip out if we're write-only
	if (fileOptions&dosfile_WRITEONLY) return fileError = resource_FileIsWriteOnly;

	// read the resource status
	if ((fileError = ReadResourceIndex( resource_num )) != resource_NoError) return fileError;

	// if the size is zero, there is no said resource
	if (index.size==0) return fileError = resource_ResourceNotFound;

	// seek to the start of the resource
	if ((fileError = Seek( index.location )) != resource_NoError) return fileError;

	// create a temporary buffer to store resource
	char*	tempbuffer = (char*) malloc (index.size+1);

	// read the resource
	fileError = DOSFile::Read(tempbuffer, index.size );

	// and decompress the resource into the buffer
	if (fileError==resource_NoError)
		DecompressResource(buffer,tempbuffer);

	// free tempbuffer
	free	(tempbuffer);

	// return our last file error
	return fileError;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		LoadAllResources
//	DESC:		malloc each data item then load each data item of a resource
//
//	-------------------------------------------------------------------------
char** BitmapResourceFile::LoadAllResources()
	{
	// get number of resources to load
	int num_resources = NumResources();

	// malloc memory for list of pointers to each resource
	grImage** datalist = (grImage**) malloc(sizeof(grImage*) * num_resources);

	int i;
	for (i=0;i<num_resources;i++)
		{
		// malloc memory for each resource
		datalist[i] = (grImage*) malloc(SizeOfDecompressedResource(i)+1);

		// read resource into memory where datalist[i] is pointing
		fileError = ReadResource (i, (char*) datalist[i]);
		}
	// return pointer to whole list
	return (char**)datalist;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		ClearAllResources
//	DESC:		free each data item of a resource
//
//	-------------------------------------------------------------------------
void BitmapResourceFile::FreeAllResources(char** datalist)
	{
	int i;
	for (i=NumResources()-1;i>=0;i--)
		// free each resource
		free(datalist[i]);

	// free whole list of pointers
	free (datalist);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		AddResource
//	DESC:		Will open a pcx file and append it to the end of the
//				resource file.
//				If you want to modify your resource file you have to delete
//				it and rewrite all the resources in, in the correct order
//
//	-------------------------------------------------------------------------
int BitmapResourceFile::AddResource(DirString* srce)
	{
//	printf("\n Source String >%s<",srce->Contents());
	int fileError = 0;
	long	pcx_size,
			buffer_size,
			resource_size;
	// check if pcx file exists then open it
	if (!dir_FileExists(srce))
		return dosfile_FileNotFound;

	// open a DOSFile
	DOSFile*		srcefile = new DOSFile(srce,dosfile_READONLY);

	// calculate pcxfile, compressed data and resource size
	pcx_size =			srcefile->FileSize();
	buffer_size =		pcx_size - PCX_HEADER_SIZE - PCX_PALETTE_SIZE;
	resource_size =		buffer_size + PCX_RESOURCE_HEADER_SIZE;

	// malloc memory for temp_pcximage where pcx file will be loaded
	PCX_image* temp_pcximage = (PCX_image*) malloc(pcx_size+1);

	// load pcx file into temp_pcximage
	fileError = srcefile->Seek (0);
	fileError = srcefile->Read ((char*)&temp_pcximage->header, PCX_HEADER_SIZE);

	fileError = srcefile->Seek (pcx_size - PCX_PALETTE_SIZE);
	fileError = srcefile->Read ((char*)&temp_pcximage->pal, PCX_PALETTE_SIZE);

	fileError = srcefile->Seek (PCX_HEADER_SIZE);
	fileError = srcefile->Read ((char*)&temp_pcximage->buffer, buffer_size);

	// create a pcx resource to save
	PCX_resource*	temp_resource = (PCX_resource*) malloc(resource_size+1);

	// convert the PCX_image to PCX_resource
	PCX_ToResource(temp_resource,temp_pcximage,buffer_size);

	// write the resource into this resourcefile
	fileError = WriteResource(NumResources(),resource_size,(char*)temp_resource);

	free(temp_resource);
	free(temp_pcximage);
	// close DOSFile
	delete srcefile;

	return fileError;
	}

