//	-------------------------------------------------------------------------
//
//	DIALOG.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	this file contains the implementation for a standard
//					Dialog Box
//
//	-------------------------------------------------------------------------

#include "dialog.hpp"
#include "stddefns.hpp"
#include "eventmgr.hpp"
#include "keydefns.hpp"

#include "stdlib.h"

//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for DialogBox
//	DESC:		The default constructor which defines the behaviour for
//				all derived DialogBoxes
//
//	PARAMS:		x1,y1,x2,y2			- the coordinates of the eventhandler
//									  (required for dialog_boxes & onscreen
//				fg,bg				- the foreground & background colors
//									  used when active, swapped when inactive
//				active_border,
//				inactive_border		- the borders used when active/inactive
//				dialog_num			- the dialog number. THIS MUST BE THE
//									  EXACT ORDER OF CONSTRUCTION THROUGH
//									  CALLS TO EVENTMGR::ADDEVENTHANDLER()!!!
//				parent_handler_num	- the parent handler number. All dialog
//									  boxes respond to a "parent" through
//									  generating a CommandEvent aimed at the
//									  parent handler.
//              options				- the type of the event handler & any
//									  required options
//				zfact				- the zfactor of the handler. When
//									  interpreting mouse & kbd events for
//									  overlapping handlers, the highest
//									  zfactor is selected to handle the input
//				retevent			- the event to return on ESC
//
//	-------------------------------------------------------------------------
DialogBox::DialogBox(int x1,int y1,int x2,int y2,uchar fg,uchar bg,
						uint active_border,uint inactive_border,
						uint dialog_num,uint parent_handler_num,
						uint options,uint zfact,uint retevent)
		: EventHandler(x1,y1,x2,y2,fg,bg,active_border,inactive_border,dialog_num,
						options|evh_DIALOG_BOX,zfact,retevent)
	{
	// by default, select button 0 (later change to user-defined)
	accessorySelected = 0;
	parent_handler = parent_handler_num;

	// caJUSTse
	isOnScreen = false;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	OpenDialog()
//	DESC:	Opens the dialog box at defined zfactor
//
//	-------------------------------------------------------------------------
void DialogBox::OpenDialog()
	{
	// skip out if dialog is already open
	if (isOnScreen) return;

	// save the image underneath
	SaveUnderneath();

	// we are now responding
	isResponding = true;
	isOnScreen = true;

	prev_handler = EVENTMGR->currentHandler;
	EVENTMGR->DeSelectEventHandler( prev_handler );

	// so draw
	Draw();

	prev_zfactor = EVENTMGR->currentZFactor;
	EVENTMGR->currentZFactor = zfactor;

	// and select our handler
	EVENTMGR->SelectEventHandler( handler_number );
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	CloseDialog()
//	DESC:	Closes the dialog box, restoring everything as on OpenDialog()
//
//	-------------------------------------------------------------------------
void DialogBox::CloseDialog()
	{
	// deselect our handler
	EVENTMGR->DeSelectEventHandler( handler_number );

	// restore what was underneath
	RestoreUnderneath( true );

	EVENTMGR->currentZFactor = prev_zfactor;

	// turn off our window
	isResponding = false;
	isOnScreen = false;

	// and select the previous handler
	EVENTMGR->SelectEventHandler( prev_handler );
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	HandleEvent(e);
//	DESC:	Checks for the ESC key which will generate the returnEvent
//
//	-------------------------------------------------------------------------
void DialogBox::HandleEvent(Event &e)
	{
	EventHandler::HandleEvent(e);

	switch (e.what)
		{
		case	ev_keyboard:		switch (e.keyboard.event)
										{
										case	ESC_UP:			CommandEvent cmd = { returnEvent, NULL };
																EVENTMGR->ModEvent(e,parent_handler,ev_command,&cmd);
																break;
										}
									break;
		}
	}
