//	-------------------------------------------------------------------------
//
//	EVENTMGR.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains defines for the EVENT MANAGER.
//					The EVENTMGR maintains internally an EVENT QUEUE, which
//					can be used through the calls GenEvent() & ModEvent()
//
//	-------------------------------------------------------------------------

#ifndef __eventmgr_hpp
#define __eventmgr_hpp

#include "event.hpp"
#include "handler.hpp"

// at most the event queue contains 32 events
// at least contains NUMBER_TIMER_TICKERS

#define EVENTMGR_DEFAULT_QUEUE_SIZE 		32

// at most the event manager can contain 64 event handlers (more == spaghetti ???)

#define EVENTMGR_MAX_EVENT_HANDLERS		64


// THE EVENT MANAGER CLASS

class EventMgr
	{
	private:
		// INTERNAL storage of the event queue (for speed!!!)
		int		nextPending;						// the next pending event
		int 	lastPending;						// the last pending event
		Event	queue[EVENTMGR_DEFAULT_QUEUE_SIZE];	// the event queue

		// INTERNAL storage of the event handlers (for speed!!!)
		EventHandler* handler[EVENTMGR_MAX_EVENT_HANDLERS];	// event handlers

		// (internal) Generic HandleEvent which handles the next pending event
		void HandleEvent(Event& e);

		// storage for saved currentHandler
		uint	saveHandler;

		// a flag to say whether or not to disable onscreen objects
		bool	eventHandlersDisabled;

	public:
		EventMgr();		// initialize the event manager
		~EventMgr();  	// terminate the event manager (as well as all handlers)

		uint	currentHandler;		// the currently selected handler
		int		currentZFactor;		// the current z dimension

		int		numEventsPending;	// how many events are pending
		int		numEventHandlers;	// the number of event handlers

		bool 	isExecuting;		// true if executing. Set to false to stop
									// Execute()'ing

		bool 	isActive;  			// true if EventMgr is active
		bool 	eventIsPending; 	// true if an event is pending in the queue
		bool 	eventQueueIsFull;  	// true if event queue is completely full
		bool 	eventHandlersAtMax;	// true if cannot have any more event handlers

		void 	SelectNextHandler();		// select the next handler (internal)
		void 	SelectPrevHandler();		// select the previous handler (internal)
		void	SelectEventHandler( uint hand );	// select a specified handler
		void	DeSelectEventHandler( uint hand );	// deselect a specified handler
		int		WhichHandlerSelected(grPoint& mouse_pos);	// which handler selected?

		// manually tells all onscreen handlers to draw
		void	DrawALLHandlers();

		// scans the event queue, returning true if the specified event is in the queue
		bool	EventInQueue(uint what,void* e);

		// Generate an event, placing it at the end of the event queue
		bool 	GenEvent(uint where, uint what, void* data);

		// Generate an event, placing it at the start of the event queue
		bool 	InsEvent(uint where, uint what, void* data);

		// Modify the current event!
		void 	ModEvent(Event& event,uint where, uint what, void* data);

		// clear a specific event
		void 	ClearEvent(Event& e);

		// clear all pending events
		void 	ClearALLEvents();

		// add an new event handler which (automatically) can then handle events
		EventHandler* AddEventHandler(EventHandler* handler);

		// remove the most recently Add'ed event handler
		void 	DelEventHandler();

		// remove ALL of the event handlers!
		void	DelALLEventHandlers();

		// the main loop of the event handling mechanism!!!!!
		void 	Execute();
	};

// globally declare the event manager!!!

extern EventMgr* EVENTMGR;

// global functions to Install/Remove the EVENTMGR

bool eventmgr_InstallEVENTMGR();
void eventmgr_RemoveEVENTMGR();

// and make it easier to access

//#define	GenEvent			EVENTMGR->GenEvent
//#define ModEvent			EVENTMGR->ModEvent
//#define ClearEvent			EVENTMGR->ClearEvent
//#define AddEventHandler		EVENTMGR->AddEventHandler
//#define DelEventHandler		EVENTMGR->DelEventHandler


#endif