//	-------------------------------------------------------------------------
//
//	FILES.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file defines the File object which is used to
//					access a pseudo-random-access file of variable-size
//					records.
//
//	-------------------------------------------------------------------------

#ifndef __files_hpp
#define __files_hpp

#include "stddefns.hpp"
#include "strings.hpp"

#include "stdio.h"
#include "io.h"

// error messages for the file object
#define file_NoError			0
#define file_FileNotFound		-1
#define file_InvalidRecord		-2
#define file_NotEnoughMem		-3
#define file_DiffRecordSize		-4
#define file_RecordNotDefined	-5
#define file_InvalidOption		-6
#define file_FileNotOpen		-7
#define file_FileIsOpen			-8
#define file_FileIsClosed		-9
#define file_BadAdjustSize		-10

// options used when opening the file
#define file_READWRITE		0x0001
#define file_READONLY		0x0002	// read-only access
#define file_WRITEONLY		0x0004	// write-only access
#define file_APPEND			0x0008	// append existing file
#define file_CREATE			0x0010	// create new file (overwrites existing file)

// options for use in directory functions (eg. FilesInDir/ReadDirectory)
#define dir_DirectoriesOnly	0x01
#define dir_FilesOnly		0x02
#define dir_AllFiles        0xFF


#define DIRSTRING_DEFAULT_LENGTH	256

class DirString : public String
	{
	private:
	protected:
	public:
		DirString(const char* st) : String(st,DIRSTRING_DEFAULT_LENGTH) {}
		~DirString() {}

		// for setting appropriate bits of a DirString
		void SetFileName(const char* st);
		void SetDrive(const char* st);
		void SetPath(const char* st);
		void SetDirectory(const char* st);

		// note: Contents() is used to return the full string....
		String* GetFileName();	// returns the filename only. 	eg. "TEST.FIL"
		String* GetDrive();		// returns the drive only. 		eg. "D:"
		String*	GetPath();		// returns the path only.		eg. "\GAME\BMED\"
		String*	GetDirectory();	// returns the directory only.	eg.	"BMED"
	};

class File
	{
	private:
		DirString* tmpFileName;	// temporary file name for DELETE, INSERT, etc.
		int  fileError; 	 	// error value set by all file routines
		int  filehandle;		// the handle for our file...
		long record_pointer;	// the last-seeked record number
		long record_size;		// size of last-seeked record
		uint file_options;		// the options for the file
		bool created_str;		// we created DirString

	protected:
	public:
		bool FileIsOpen;		// whether or not the file is open
		DirString* FileName;		// name of the File

		File(DirString* pathname,uint options);	// defines the File with options
		File(const char* pathname,uint options);// defines the File with options
		~File();								// close the file

		int  Seek(long recnum);					// seek record # recnum
		long Recsize(long recnum);				// find size of record # recnum
		void SetFileOptions(DirString* filename,uint options);	// set the file options once NEWd
		int  Open();							// open the file
		int  Close();							// close the file
		int  Read(long recnum, void* data);		// read from file
		int  Write(long recnum, void* data, long size); // write to file..
		int  Delete(long recnum);        		// delete a record
		int  Insert(long recnum,long size);		// insert a record
		int  Error() { return fileError; }		// returns result of last operation
		bool Eof() { return eof(filehandle); }	// return true if at End-of-file
	};

/*
bool fileExists(DirString* pathname);	// returns true if file exists
int FilesInDir(uint attrib);					// returns # of files in directory
void ReadDirectory(String* dirstring,uint attrib);	// builds a string containing files of type 'attrib'.
													// eg. "A.DAT/B.DAT/C.DAT/ETC.ETC"
void GetDriveNames(String* diskstr);	// builds a string of current available drives
										// eg. "A:/B:/C:/D:"
void GetCurrentPath(String* pathstr);	// builds a string containing the current path.
										// eg. "D:\GAME\BMED\"
*/
#endif