//	-------------------------------------------------------------------------
//
//	GRAPHICS.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains the base level graphics functions used
//					by all of the game development library.
//
//	-------------------------------------------------------------------------

#include "graphics.hpp"
#include "mouse.hpp"
#include "realmode.hpp"
#include "fonts.hpp"

#include "conio.h"
#include "mem.h"
#include "dos.h"
#include "stdio.h"
#include "stdarg.h"
#include "stdlib.h"

//#define __debugging	// used for disabling/enabling error messages

#define 	PAL_MASK		0x03C6
#define		PAL_READ		0x03C7
#define		PAL_WRITE		0x03C8
#define		PAL_DATA		0x03C9

uchar	gr_HIGHLIGHT  = 30;
uchar	gr_LOWLIGHT   = 24;
uchar	gr_BACKGROUND = 7;

uchar 	FGCOLOR = 8;	// darkgray
uchar	BGCOLOR = 7;	// lightgray
uchar	LIGHTCOLOR = 30;
uchar	DARKCOLOR = 24;

grPalette*  grPALETTE;			// the current palette
grRect 		grFULLSCREEN;		// the full screen as a rectangle
float 		red_step[256];		// used by grFadeOutPalette & grFadeInPalette
float 		green_step[256];	// ditto
float 		blue_step[256];		// ditto
uint 		grMaxx;				// the maximum x coordinate
uint 		grMaxy;				// the maximum y coordinate
bool 		grMouseIsHidden;	// whether or not the mouse is currently hidden
uint		grLastVideoMode;	// the last video mode (saved on startup)
uchar		grDrawColor;		// the current draw color

// the following are used by the grSwitchRead/WriteBank() functions

uint		current_read_bank = 0xFFFFFFFF;  // the current read bank
uint		current_write_bank = 0xFFFFFFFF; // the current write bank
bool		separate_write_bank = false;	 // are the banks separate???
uint		bank_offset = 0;	// the bank offset for different page numbers

// bits and pieces for the font engine

#define		MAX_LOADABLE_FONTS		20
const int	listsize = 				95;
int			lastfont = 				font_NoFontsLoaded;

char**				currentfontPointer = NULL;
char**				fontPointers [MAX_LOADABLE_FONTS];
FontResourceFile*	fontResource [MAX_LOADABLE_FONTS];

// the following are used by the grSetPaletteEntry(), etc. functions

int GR_GRID_XOFFSET = 0;
int GR_GRID_YOFFSET = 0;
int GR_GRID_XBOXES = 0;
int GR_GRID_YBOXES = 0;
int GR_GRID_BOXWIDTH = 0;
int GR_GRID_BOXHEIGHT = 0;
uchar GR_GRID_COLOR = 8;

// the VESA information structure

struct VESA_INFO_BLOCK
	{
	BYTE 	V[4];
	WORD  	version,
			oem_ofs,
			oem_seg;
	DWORD 	capabilities;
	WORD	modes_ofs,
			modes_seg,
			memblocks;
	BYTE 	wasted[400];	// extra large due to bugs in some VESA BIOS's
	};

// the VESA mode information structure

struct VESA_MODE_BLOCK
	{
	WORD 	vesainfo;
	BYTE	winaatts,winbatts;
	WORD	grain,
			size,
			aseg,
			bseg;
	void far* bankfunc;
	WORD	bytesperrow,
			xres,
			yres;

			// note: the following are not used in VESA modes except for
			//		 supporting OEM modes... (ie. 0x00 - 0xFF)
	BYTE	xchars,ychars,bitplanes,bppixel,numbanks,memmodel,
			banksize,planes,reserved,redmask,redpos,greenmask,
			greenpos,bluemask,bluepos,rsvdmask,rsvdpos,dcinfo;

	BYTE 	wasted[216];
	};

VESA_INFO_BLOCK 	VESAinfo;	// storage of vesa information
VESA_MODE_BLOCK		VESAmodes;	// storage of vesa mode information

WORD	VESAreadwindow; 		// either window A or window B
WORD	VESAwritewindow;		// either window A or window B
bool	VESA_usingsamewindow;	// VESAreadwindow is same as VESAwritewindow

#define	video_addr		0x000A0000	// video memory absolute start location

// =========================================================================
//
// VESA DRIVER FUNCTIONS
//
// =========================================================================
//
// ----- the following are used by grInitVesa(), grExitVesa() & graphics
//		 primitives such as grPutPixel(), grPutImage() etc. These should
//		 not be used unless you know what you are doing!!!

// returns true if VESA 1.2 compatible bios is currently available
bool grVesaBiosInstalled(uint mode);

// returns true if requested VESA mode is available
bool grVesaModeAvailable(uint mode);

// returns true if could set requested VESA mode
bool grSetVesaMode(uint mode);

// resets the video device to mode on startup
void grResetMode(uint mode = grLastVideoMode);

// switch write bank for VESA compatibles
void grSwitchWriteBank(uint newbank);

// switch read bank for VESA compatibles
void grSwitchReadBank(uint newbank);

// 	--------------------------------------------------------------------------
//
// 	FUNC:	grInitVesa()
// 	DESC:	Initialises the graphics display, installing the mouse driver,
//			etc.
//
// 	PARAMS:	mode	- the video mode to set.
//						0x0100H = 640x400x256 packed
//						0x0101H = 640x480x256 packed
//						etc.
//
// 	--------------------------------------------------------------------------
bool grInitVesa(uint mode)
	{
	// attempt to set specified vesa mode
	if (grSetVesaMode(mode))
		{
		// we were successful, initialize mouse, etc.
		grPALETTE = grCreatePalette(256);
		if (grPALETTE==NULL)
			{
			grResetMode();
			printf("\nERROR: could not create grPALETTE");
			exit(1);
			}

		grInitMouse();
		return true;
		}
	else
		return false;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grExitVesa
//	DESC:	terminates the display, resetting as before grInitVesa
//
// --------------------------------------------------------------------------
void grExitVesa()
	{
	// remove global palette info
	free( grPALETTE );

	// terminate mouse driver
	grExitMouse();

	// and reset video mode
	grResetMode();
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grClearVideo
//	DESC:	clears the display, optionally setting the color to 'color'
//
// --------------------------------------------------------------------------
void grClearVideo(uchar color)
	{
	// create DWORD mask for color
	DWORD fill = color + color<<8 + color<<16 + color<<24;

	DWORD* p;
	ulong banknum;
	ulong dwords;

//	grHideMouse();

	// for all available banks...
	for (banknum=0;banknum<8;banknum++)
		{
		grSwitchWriteBank(banknum);

		// blit DWORDS...
		p = (DWORD*) video_addr;
		for (dwords=0;dwords<16384;dwords++)
			{
			*p = fill;
			p++;
			}
		}

//	grShowMouse();
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grVesaBiosInstalled()
//	DESC:	checks to see if there is a VESA 1.2 driver installed.
//			If there is a driver, we then get all required vesa information
//			and return true. Otherwise we return false.
//
//	NOTE:	if (#defined __debugging) display vesa information
//
// --------------------------------------------------------------------------
bool grVesaBiosInstalled(uint mode)
	{
	SREGS sregs;
	REGS regs;

	// clear our register variables
	memset(&sregs,0,sizeof(sregs));
	memset(&regs,0,sizeof(regs));

	// set up for VESA bios call 0x4F03 - Get Video Mode
	regs.w.ax = 0x4F03;
	RM_int86(0x10,&regs,&regs);

	// if we couldn't get video mode, skip out!!!!
	if (regs.w.ax!=0x004F) return false;

	// save mode for ResetMode()
	grLastVideoMode = regs.w.bx;

	// set up video bios call 0x4F00 - Get SuperVGA Information
	memset(&sregs,0,sizeof(sregs));
	memset(&regs,0,sizeof(regs));

	// create DOS memory to store the vesa mode information
	short selector,segment;
	char far* dos_info = (char far*) RM_malloc(&selector,&segment,sizeof(VESAinfo));

	// make ES:(E)DI point to our RM_malloc result
	sregs.es = segment;
	regs.x.edi = 0;
	regs.w.ax = 0x4F00;

	// and get the VESA bios information
	RM_int86x(0x10,&regs,&regs,&sregs);

	// save info into parm
	_fmemcpy(&VESAinfo,dos_info,sizeof(VESAinfo));

	// and clean up our dos memory usage
	RM_free(selector);

#ifdef __debugging
	printf("\nVESA INFORMATION");
	printf("\nV              = >%c%c%c%c<",VESAinfo.V[0],VESAinfo.V[1],VESAinfo.V[2],VESAinfo.V[3]);
	printf("\nversion        = %u",VESAinfo.version);
	printf("\noem_ofs        = %x",VESAinfo.oem_ofs);
	printf("\noem_seg        = %x",VESAinfo.oem_seg);
	printf("\ncapabilities   = %l",VESAinfo.capabilities);
	printf("\nmodes_ofs      = %x",VESAinfo.modes_ofs);
	printf("\nmodes_seg      = %x",VESAinfo.modes_seg);
	printf("\nmemblocks      = %d",VESAinfo.memblocks);
	printf("\n");
	getch();
#endif

	// if we didn't get the right information, skip out
	if ((regs.w.ax!=0x004F) || (strncmp("VESA",(char*)&VESAinfo.V[0],4)!=0)) return false;

	// clear regs
	memset(&regs,0,sizeof(regs));
	memset(&sregs,0,sizeof(sregs));

	// set up VESA bios call 0x4F01 - Get SuperVGA Mode Information
	char far* dos_modes = (char far*) RM_malloc(&selector,&segment,sizeof(VESAmodes));

	// make ES:(E)DI point to our RM_malloc result
	sregs.es = segment;
	regs.x.edi = 0;
	regs.w.ax = 0x4F01;
	regs.w.cx = mode;

	// and get the VESA mode information
	RM_int86x(0x10,&regs,&regs,&sregs);

	// save info into parm
	_fmemcpy(&VESAmodes,dos_modes,sizeof(VESAmodes));

	// and clean up our dos memory usage
	RM_free(selector);

	// skip out if invalid call
	if (regs.w.ax!=0x004F) return false;

#ifdef __debugging
	printf("\nVESA MODE INFORMATION:");
	printf("\nvesainfo         = %u",VESAmodes.vesainfo);
	printf("\nwinaatts         = %u",VESAmodes.winaatts);
	printf("\nwinbatts         = %u",VESAmodes.winbatts);
	printf("\ngrain            = %u",VESAmodes.grain);
	printf("\nsize             = %u",VESAmodes.size);
	printf("\naseg             = %u",VESAmodes.aseg);
	printf("\nbseg             = %u",VESAmodes.bseg);
	printf("\nbankfunc         = %P",VESAmodes.bankfunc);
	printf("\nbytesperrow      = %u",VESAmodes.bytesperrow);
	if (mode<0x0100)
		{
		printf("\nxres             = %u",VESAmodes.xres);
		printf("\nyres             = %u",VESAmodes.yres);
		printf("\nxchars           = %u",VESAmodes.xchars);
		printf("\nychars           = %u",VESAmodes.ychars);
		printf("\nbitplanes        = %u",VESAmodes.bitplanes);
		printf("\nbppixel          = %u",VESAmodes.bppixel);
		printf("\nnumbanks         = %u",VESAmodes.numbanks);
		printf("\nmemmodel         = %u",VESAmodes.memmodel);
		printf("\nbanksize         = %u",VESAmodes.banksize);
		printf("\nplanes           = %u",VESAmodes.planes);
		printf("\nredmask          = %u",VESAmodes.redmask);
		printf("\nredpos           = %u",VESAmodes.redpos);
		printf("\ngreenmask        = %u",VESAmodes.greenmask);
		printf("\ngreenpos         = %u",VESAmodes.greenpos);
		printf("\nbluemask         = %u",VESAmodes.bluemask);
		printf("\nbluepos          = %u",VESAmodes.bluepos);
		printf("\ndcinfo           = %u",VESAmodes.dcinfo);
		}
	printf("\n");
	getch();
#endif

	// vesa bios is available
	return true;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grVesaModeAvailable()
//	DESC:	returns true if the specified mode is available.
//
// --------------------------------------------------------------------------
bool grVesaModeAvailable(uint mode)
	{
	return true;
/*	unsigned int far *listptr;

	// check for 640 * 480 * 256 mode
	for (listptr = (unsigned int *) MK_FP(parm.modes_seg,parm.modes_ofs);
		 *listptr != 0xFFFF && *listptr != mode; listptr++);

	// return false if mode is not available
	if (*listptr == 0xFFFF)
		return false;

	// specified mode is available
	return true;
*/	}

// --------------------------------------------------------------------------
//
//	FUNC:	grSetVesaMode()
//	DESC:	Checks whether the driver/hardware can supply whats required.
//			If it can, set the video mode & return true.
//
// --------------------------------------------------------------------------
bool grSetVesaMode(uint mode)
	{
#ifdef __debugging
	printf("\nchecking for VESA BIOS");
#endif
	// return if there is no VESA bios installed in system
	if (!grVesaBiosInstalled(mode)) return false;

#ifdef __debugging
	printf("\nchecking for VESA MODE");
#endif
	// also return if we could get mode information on requested mode
	if (!grVesaModeAvailable(mode)) return false;

	// vesainfo (bit0) == 1
	if ((VESAmodes.vesainfo & 1)==1)
		{
		// attributes are as follows:
		//		bit 0	- window exists
		//		bit 1	- window is readable
		//		bit 2	- window is writeable

		// get the read & write segments
		uint writeseg = VESAmodes.aseg;
		uint readseg = VESAmodes.aseg;

		// both segments _must_ be the same
		if (readseg!=writeseg)
			{
			grResetMode();
			printf("\nERROR: read/write windows do not overlap.");
			printf("\n       For this program to work, they must.");
			getch();
			return false;
			}

		// if window B is writable
		if ((VESAmodes.winbatts & 5)==5)
			{
			VESAwritewindow = 0x0001;	// use window B
#ifdef __debugging
			printf("\nWRITE WINDOW == B\n");
			getch();
#endif
			}
		else
			// if window A exists & is writeable
			if ((VESAmodes.winaatts & 4)==4)
				{
				VESAwritewindow = 0x0000;	// use window A
#ifdef __debugging
				printf("\nWRITE WINDOW == A\n");
				getch();
#endif
				}
			else
				// error???
				{
				grResetMode();
				printf("\nERROR: could not find an appropriate write bank.");
				getch();
				return false;
				}

		// if window A is readable.
		if ((VESAmodes.winaatts & 2)==2)
			{
			VESAreadwindow = 0x0000;	// use window A
#ifdef __debugging
			printf("\nREAD WINDOW == A\n");
			getch();
#endif
			}
		else
			// if window B exists & is readable
			if ((VESAmodes.winbatts & 3)==3)
				{
				VESAreadwindow = 0x0001;	// use window B
#ifdef __debugging
				printf("\nREAD WINDOW == B\n");
				getch();
#endif
				}
			else
				// error???
				{
				grResetMode();
				printf("\nERROR: could not find an appropriate read bank.");
				getch();
				return false;
				}

		VESA_usingsamewindow = (VESAreadwindow==VESAwritewindow);

		// this program only works with 64K banks
		if (VESAmodes.size!=64)
			{
			grResetMode();
			printf("\nERROR: Your video driver/hardware configuration");
			printf("\n       does not supply 64k graphic windows.");
			getch();
			return false;
			}

		// set up VESA bios call 0x4F02 - Set Video Mode
		REGS regs;
		memset(&regs,0,sizeof(regs));

		regs.w.ax = 0x4F02;
		regs.w.bx = mode;
		RM_int86(0x10,&regs,&regs);

		// if we couldn't set vesa mode, return error
		if (regs.w.ax!=0x004F)
			{
			grResetMode();
			printf("ERROR:: COULD NOT SET VESA MODE!!!!");
			return false;
			}

		// set up VESA bios call 0x4F06 - Get/Set Logical Scan Line Length
		memset(&regs,0,sizeof(regs));

		regs.w.ax = 0x4F06;
		regs.w.bx = 0x0000;
		regs.w.cx = 1024;	// length of virtual scan line - 1024 pixels
		RM_int86(0x10,&regs,&regs);

		// skip out if not successful
		if (regs.w.ax!=0x004F)
			{
			grResetMode();
			printf("\nERROR: Could not set the virtual scan line to 1024 pixels");
			getch();
			return false;
			}
		}
	else
		{
		grResetMode();
		printf("\nERROR: VESA attributes bit 0 is not set!!!");
		getch();
		return false;
		}

	// clear our video display
	grClearVideo();

	// and we've done it!!!
#ifdef __debugging
	printf("\nVESA MODE WAS SET CORRECTLY");
#endif
	return true;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grResetMode()
//	DESC:	Sets the video mode to mode # 'mode'. By default it resets
//			the mode to that on startup...
//
// --------------------------------------------------------------------------
void grResetMode(uint mode)
	{
	REGS regs;
	memset(&regs,0,sizeof(regs));

	// and set the video mode
	regs.w.ax = 0x4F02;
	regs.w.bx = mode;
	RM_int86(0x10,&regs,&regs);
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grSwitchWriteBank()
//	DESC:	Sets the write bank for screen writing routines. If the
//			read/write banks are separate, this uses window B, otherwise
//			it uses window A.
//
//	RETURNS:	on success, the global variable 'current_write_bank' is set
//
// --------------------------------------------------------------------------
void grSwitchWriteBank(uint newbank)
	{
	REGS regs;
	memset(&regs,0,sizeof(regs));

	// set up for VESA mode 0x4F05	- CPU video memory control
	regs.w.ax = 0x4F05;
	regs.w.bx = VESAwritewindow;
	regs.w.dx = newbank + bank_offset;	// bank_offset is global, set by setpage();
	RM_int86(0x10,&regs,&regs);

	if (regs.w.ax!=0x004F)		// iff !successful
		{
		grResetMode();
		printf("\nERROR: could not set WRITE bank #%d",newbank);
		printf("\n");
		getch();
		exit(1);
		}
	else
		{
		current_write_bank = newbank;
		if (VESA_usingsamewindow) current_read_bank = newbank;
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grSwitchReadBank()
//	DESC:	Sets the read bank for screen reading routines. This uses
//			window A always.
//
//	RETURNS:	on success, the global variable 'current_read_bank' is set
//
// --------------------------------------------------------------------------
void grSwitchReadBank(uint newbank)
	{
	REGS regs;
	memset(&regs,0,sizeof(regs));

	// set up for VESA mode 0x4F05	- CPU video memory control
	regs.w.ax = 0x4F05;
	regs.w.bx = VESAreadwindow;			// select appropriate window
	regs.w.dx = newbank + bank_offset;	// bank_offset is global, set by setpage();
	RM_int86(0x10,&regs,&regs);


	if (regs.w.ax!=0x004F)		// iff !successful
		{
		grResetMode();
		printf("\nERROR: could not set READ bank #%d",newbank);
		printf("\n");
		getch();
		exit(1);
		}
	else
		{
		current_read_bank = newbank;
		if (VESA_usingsamewindow) current_write_bank = newbank;
		}
	}

// =========================================================================
//
//	VIRTUAL DISPLAY FUNCTIONS
//
// =========================================================================
//
//	FUNC:		grCreateVirtual()
// 	DESC:		Creates a virtual display buffer, generally used to store a
//				complex image which is created with a number of different
//				writes. This image can be placed on the screen by calling
//				grBlitVirtual().
//
//	RETURNS:	On success, a pointer to the virtual buffer is returned
//				On error, returns NULL.
//
// --------------------------------------------------------------------------
grVirtualBuffer* grCreateVirtual(uint x1,uint y1,uint x2,uint y2)
	{
	// malloc memory
	grVirtualBuffer*p = (grVirtualBuffer*) malloc(grImageSize(x1,y1,x2,y2));

	// and create a header
	p->hdr.width  = (x2-x1+1);
	p->hdr.height = (y2-y1+1);

	return (grVirtualBuffer*) p;
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDestroyVirtual()
// 	DESC:		Deletes a virtual buffer created with grCreateVirtual().
//
// --------------------------------------------------------------------------
void grDestroyVirtual(grVirtualBuffer* virt)
	{
	free((char*) virt);
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grClearVirtual()
// 	DESC:		Clears a virtual buffer to a specified color. grBLACK by default
//
// --------------------------------------------------------------------------
void grClearVirtual(grVirtualBuffer* virt,uchar color)
	{
	uint x1 = 0;
	uint x2 = virt->hdr.width;
	uint y1 = 0;
	uint y2 = virt->hdr.height;
	uint x,y;

	WORD	wfill = color + (color<<8);
	DWORD	dfill = wfill + (wfill<<16);

	ulong virt_addr 	= (ulong) virt + sizeof(grImageHeader);

	uint width 			= x2 - x1 + 1;			// total number of bytes to copy
	uint dwords_to_copy = width >> 2;			// dwords to copy == bytes / 4
	uint words_to_copy  = 0;//bytesleft & 2;		// if word-aligned, copy 1 word
	uint bytes_to_copy 	= 0;//bytesleft & 3;		// if byte-aligned, copy 1 byte

	switch (width&3)
		{
		case 0: break;
		case 1: bytes_to_copy = 1; break;
		case 2: words_to_copy = 1; break;
		case 3: words_to_copy = 1;
				bytes_to_copy = 1; break;
		}

	// calc destination inside virtual buffer
	DWORD* dest = (DWORD*) virt_addr;

	// splat all DWORDS across
	for (x=0;x<dwords_to_copy;x++)
		{
		*dest = dfill;
		dest++;
		}

	// if we need to copy a WORD
	if (words_to_copy>0)
		{
		WORD* wdest = (WORD*) dest;
		*wdest = wfill;
		wdest++;

		// we may need to copy a BYTE as well
		if (bytes_to_copy>0)
			{
			BYTE* bdest = (BYTE*) wdest;
			*bdest = color;
			}
		}
	else
		// we don't need to copy a word
		// but we may still need to copy a BYTE
		{
		if (bytes_to_copy>0)
			{
			BYTE* bdest1 = (BYTE*) dest;
			*bdest1 = color;
			}
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grScrollVirtual()
// 	DESC:		Scrolls a virtual buffer by dx???,dy??? pixels
//
// --------------------------------------------------------------------------
void grScrollVirtual(grVirtualBuffer* virt,int dx,int dy)
	{
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grPutPixelVirtual()
// 	DESC:		sets a specified pixel in virtual to 'color'
//
// --------------------------------------------------------------------------
void grPutPixelVirtual(grVirtualBuffer* virt,uint x,uint y,uchar color)
	{
	// clip...
	CLIP(x,y,0,0,virt->hdr.width,virt->hdr.height);

	// and put the pixel!!
	virt->data[y*virt->hdr.width+x] = color;
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grGetPixelVirtual()
// 	DESC:		gets a specified pixel in virtual & returns it
//
// --------------------------------------------------------------------------
uchar grGetPixelVirtual(grVirtualBuffer* virt,uint x,uint y)
	{
	// clip...
	CLIP(x,y,0,0,virt->hdr.width,virt->hdr.height);

	// and get the pixel!!
	return virt->data[y*virt->hdr.width+x];
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grPutImageVirtual()
// 	DESC:		puts an image into a virtual buffer starting at x1,y1
//
// --------------------------------------------------------------------------
void grPutImageVirtual(grImage* virt,uint x1,uint y1,uint x2,uint y2,grImage* image)
	{
	ulong virt_addr = (ulong) virt + sizeof(grImageHeader);
	ulong image_addr= (ulong) image+ sizeof(grImageHeader);

	// make sure coords are clipped!!!
	CLIP(x1,y1,0,0,virt->hdr.width,virt->hdr.height);
	CLIP(x2,y2,0,0,virt->hdr.width,virt->hdr.height);

	uint virt_width  = virt->hdr.width;
	uint image_width = x2-x1+1;

	uint dwords_to_copy = image_width >> 2;
	uint words_to_copy  = 0;
	uint bytes_to_copy  = 0;
	uint x,y;

	switch (image_width&3)
		{
		case 0: break;
		case 1: bytes_to_copy = 1; break;
		case 2: words_to_copy = 1; break;
		case 3: words_to_copy = 1;
				bytes_to_copy = 1; break;
		}

	// blit the image

	DWORD* dest;
	DWORD* srce;
	for (y=y1;y<=y2;y++)
		{
		dest = (DWORD*) (virt_addr + (y * virt_width) + x1); // same as relative_y * 1024

		// and our offset into the picture we're working with
		srce = (DWORD*) (image_addr);
		// adjust our index into the picture to reflect line being drawn
		image_addr += image_width;

		// splat all DWORDS across
		for (x=0;x<dwords_to_copy;x++)
			{
			*dest = *srce;
			srce++;
			dest++;
			}

		// if we need to copy a WORD
		if (words_to_copy>0)
			{
			WORD* wdest = (WORD*) dest;
			WORD* wsrce = (WORD*) srce;
			*wdest = *wsrce;
			wdest++;
			wsrce++;

			// we may need to copy a BYTE as well
			if (bytes_to_copy>0)
				{
				BYTE* bdest = (BYTE*) wdest;
				BYTE* bsrce = (BYTE*) wsrce;
				*bdest = *bsrce;
				}
			}
		else
			// we don't need to copy a word
			// but we may still need to copy a BYTE
			{
			if (bytes_to_copy>0)
				{
				BYTE* bdest1 = (BYTE*) dest;
				BYTE* bsrce1 = (BYTE*) srce;
				*bdest1 = *bsrce1;
				}
			}
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grOverlayImageVirtual()
// 	DESC:		overlays an image into a virtual buffer starting at x1,y1
//
// --------------------------------------------------------------------------
void grOverlayImageVirtual(grVirtualBuffer* virt,uint x1,uint y1,uint x2,uint y2,grImage* image)
	{
	ulong virt_addr = (ulong) virt + sizeof(grImageHeader);
	ulong image_addr= (ulong) image+ sizeof(grImageHeader);

	// make sure coords are clipped!!!
	CLIP(x1,y1,0,0,virt->hdr.width,virt->hdr.width);
	CLIP(x2,y2,0,0,virt->hdr.width,virt->hdr.width);

	uint virt_width  = virt->hdr.width;
	uint image_width = image->hdr.width;
	uint x,y;

	// blit the image

	BYTE* dest;
	BYTE* srce;
	BYTE  tmp_srce;

	for (y=y1;y<=y2;y++)
		{
		dest = (BYTE*) (virt_addr + (y * virt_width) + x1); // same as relative_y * 1024

		// and our offset into the picture we're working with
		srce = (BYTE*) (image_addr);
		// adjust our index into the picture to reflect line being drawn
		image_addr += image_width;

		// splat all BYTES across
		for (x=0;x<image_width;x++)
			{
			tmp_srce = *srce;	// quicker than if (*srce) *dest=*srce; due to video bus...
			if (tmp_srce) *dest = tmp_srce;
			srce++;
			dest++;
			}

		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDrawLineVirtual()
// 	DESC:		draw a line in a virtual buffer from x1,y1 to x2,y2
//
// --------------------------------------------------------------------------
void grDrawLineVirtual(grVirtualBuffer* virt,uint x1, uint y1, uint x2, uint y2, uchar color)
	{
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDrawVLineVirtual()
// 	DESC:		draw a vertical line in a virtual buffer from x,y1 to x,y2
//
// --------------------------------------------------------------------------
void grDrawVLineVirtual(grVirtualBuffer* virt,uint x,uint y1,uint y2,uchar color)
	{
	CLIP(x,y1,0,0,virt->hdr.width,virt->hdr.height);
	CLIP(x,y2,0,0,virt->hdr.width,virt->hdr.height);

	uint yy1 = y1;
	uint yy2 = y2;

	if (yy1==yy2) { grPutPixelVirtual(virt,x,yy1,color); return; }
	if (yy1>yy2) { yy1 = y2; yy2 = y1; }

	ulong virt_addr = (ulong) virt + sizeof(grImageHeader);
	uint virt_width = virt->hdr.width;

	for (uint y=yy1;y<=yy2;y++)
		{
		BYTE* dest = (BYTE*) (virt_addr + (y * virt_width) + x);
		*dest = color;
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDrawHLineVirtual()
// 	DESC:		draw a horizontal line in a virtual buffer from x1,y to x2,y
//
// --------------------------------------------------------------------------
void grDrawHLineVirtual(grVirtualBuffer* virt,uint y,uint x1,uint x2,uchar color)
	{
	CLIP(x1,y,0,0,virt->hdr.width,virt->hdr.height);
	CLIP(x2,y,0,0,virt->hdr.width,virt->hdr.height);

	if (x1==x2) { grPutPixelVirtual(virt,x1,y,color); return; }

	ulong virt_addr = (ulong) virt + sizeof(grImageHeader);
	uint virt_width = virt->hdr.width;

	uint width = x2 - x1 + 1;
	uint dwords_to_copy = width >> 2;
	uint words_to_copy = 0;
	uint bytes_to_copy = 0;
	uint x;
	WORD wcolor = color + (color<<8);
	DWORD dcolor = wcolor + (wcolor<<16);

	switch (width&3)
		{
		case 0: break;
		case 1: bytes_to_copy = 1; break;
		case 2: words_to_copy = 1; break;
		case 3: words_to_copy = 1;
				bytes_to_copy = 1; break;
		}

	DWORD* dest = (DWORD*) (virt_addr + (y * virt_width) + x1);
	// splat all DWORDS across
	for (x=0;x<dwords_to_copy;x++)
		{
		*dest = dcolor;
		dest++;
		}

	// if we need to copy a WORD
	if (words_to_copy>0)
		{
		WORD* wdest = (WORD*) dest;
		*wdest = wcolor;
		// we may need to copy a BYTE as well
		if (bytes_to_copy>0)
			{
			BYTE* bdest = (BYTE*) wdest;
			*bdest = color;
			}
		}
	else
		// we don't need to copy a word
		// but we may still need to copy a BYTE
		{
		if (bytes_to_copy>0)
			{
			BYTE* bdest1 = (BYTE*) dest;
			*bdest1 = color;
			}
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDrawRectVirtual()
// 	DESC:		draws a rectangle in a virtual buffer from x1,y1 to x2,y2
//
// --------------------------------------------------------------------------
void grDrawRectVirtual(grVirtualBuffer* virt,uint x1, uint y1, uint x2, uint y2, uchar color)
	{
	grDrawVLineVirtual(virt,x1,y1,y2,color);
	grDrawVLineVirtual(virt,x2,y1,y2,color);
	grDrawHLineVirtual(virt,y1,x1,x2,color);
	grDrawHLineVirtual(virt,y2,x1,x2,color);
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDrawFilledRectVirtual()
// 	DESC:		draws a filled rectangle in a virtual buffer from x1,y1 to x2,y2
//
// --------------------------------------------------------------------------
void grDrawFilledRectVirtual(grVirtualBuffer* virt,uint x1,uint y1,uint x2,uint y2, uchar color)
	{
	CLIP(x1,y1,0,0,virt->hdr.width,virt->hdr.height);
	CLIP(x2,y2,0,0,virt->hdr.width,virt->hdr.height);

	uint x,y;

	WORD 	wfill = color + (color<<8);
	DWORD   dfill = wfill + (wfill<<16);

	DWORD* 	dest;

	ulong 	virt_addr = (ulong) virt;
	uint 	virt_width 	= virt->hdr.width;

	uint rect_width 	= x2 - x1 + 1;			// total number of bytes to copy
	uint dwords_to_copy = rect_width >> 2;			// dwords to copy == bytes / 4
	uint words_to_copy  = 0;//bytesleft & 2;		// if word-aligned, copy 1 word
	uint bytes_to_copy 	= 0;//bytesleft & 3;		// if byte-aligned, copy 1 byte

	switch (rect_width&3)
		{
		case 0: break;
		case 1: bytes_to_copy = 1; break;
		case 2: words_to_copy = 1; break;
		case 3: words_to_copy = 1;
				bytes_to_copy = 1; break;
		}

	// blit the image

	for (y=y1;y<=y2;y++)
		{
		dest = (DWORD*) (virt_addr + (y * virt_width) + x1);

		// splat all DWORDS across
		for (x=0;x<dwords_to_copy;x++)
			{
			*dest = dfill;
			dest++;
			}

		// if we need to copy a WORD
		if (words_to_copy>0)
			{
			WORD* wdest = (WORD*) dest;
			*wdest = wfill;
			wdest++;

			// we may need to copy a BYTE as well
			if (bytes_to_copy>0)
				{
				BYTE* bdest = (BYTE*) wdest;
				*bdest = color;
				}
			}
		else
			// we don't need to copy a word
			// but we may still need to copy a BYTE
			{
			if (bytes_to_copy>0)
				{
				BYTE* bdest1 = (BYTE*) dest;
				*bdest1 = color;
				}
			}
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDrawStringVirtual()
// 	DESC:		draws a string in a virtual buffer at location x,y
//
// --------------------------------------------------------------------------
void grDrawStringVirtual(grVirtualBuffer* virt,int x,int y,char* fmt,...)
	{
	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string
	grImage*character;		// current character to be displayed
	int		currentx;		// store current x for each character
	int		i;				// counter

	va_start(argptr,fmt);	// figure out the argument starting address

	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	// and display the text onto the screen
	grHideMouse();

	currentx = x;
	// display each character of buf string
	for (i=0;buf[i]!='\0';i++)
		{
		// *NOTE:: the font is stored in the same order as the ascii set
		// but the first character is the space key 32, the list is contiguous
		// after this for 95 characters. See top of file for full list
		character = (grImage*)currentfontPointer[buf[i]-32];

		// put image into the virtual buffer
		grOverlayImageVirtual(virt,currentx,y,currentx+character->hdr.width-1,
								y+character->hdr.height-1,character);

		// increment currentx by last char's image width
		currentx += character->hdr.width;
		}
	grShowMouse();

	va_end(argptr);				// to perform a normal return from this func
	}


// =========================================================================
//
//	DIRECT-TO-VIDEO FUNCTIONS
//
// =========================================================================
//
// FUNC:	grPutPixel()
// DESC:	places a pixel on the screen at coordinates x,y with color 'color'
//
// --------------------------------------------------------------------------
void grPutPixel(uint x,uint y,uchar color)
	{
	uint relative_y = y & 63;
	uint required_bank = y >> 6;

	if (required_bank!=current_write_bank) grSwitchWriteBank(required_bank);

	BYTE* dest = (BYTE*) (video_addr + (relative_y << 10) + x);
	*dest = color;
	}

//	--------------------------------------------------------------------------
//
//	FUNC:	grGetPixel()
//	DESC:	retrieves a pixel from the screen at coordinates x,y
//
//	--------------------------------------------------------------------------
uchar grGetPixel(uint x,uint y)
	{
	uint relative_y = y & 63;
	uint required_bank = y >> 6;

	if (required_bank!=current_read_bank) grSwitchReadBank(required_bank);

	BYTE* dest = (BYTE*) (video_addr + (relative_y << 10) + x);
	return *dest;
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grPutImage()
//
//	DESC:		places image onto screen.
//
//	PARAMS:		x1,y1,x2,y2	- the rectangular region to place the image in.
//				image		- a pointer to a grImage
//
//	NOTES:		This routine expects the image to be the same size or greater
//				than the size of the rectangular region specified by x1,y1,x2,y2
//				As such it clips only the x1,y1,x2,y2 bounds. Sorry, but its quicker!!!
//				(ie. If the image is 8x8 and you place it in a 16x16 region,
//				 all out of bounds bits will generally be utter crap)
//
//	USAGE:		char* pic = (char*) malloc(grImageSize(0,0,99,99));
//				grPutImage(0,0,99,99,pic);			// not clipped
//				grPutImage(200,200,249,249,pic);	// clipped
//				free(pic);
//
//	--------------------------------------------------------------------------
void grPutImage(uint x1,uint y1,uint x2,uint y2,grImage* image)
	{
	uint x,y;
	uint relative_y;
	ulong bank_num;

	DWORD* dest;
	DWORD* srce;

	ulong image_addr	= (ulong) image + sizeof(grImageHeader);

	uint width 			= x2 - x1 + 1;			// total number of bytes to copy
	uint dwords_to_copy = width >> 2;			// dwords to copy == bytes / 4
	uint words_to_copy  = 0;//bytesleft & 2;		// if word-aligned, copy 1 word
	uint bytes_to_copy 	= 0;//bytesleft & 3;		// if byte-aligned, copy 1 byte

	switch (width&3)
		{
		case 0: break;
		case 1: bytes_to_copy = 1; break;
		case 2: words_to_copy = 1; break;
		case 3: words_to_copy = 1;
				bytes_to_copy = 1; break;
		}

	// blit the image

	for (y=y1;y<=y2;y++)
		{
		// calc address 0x000A0000 -> 0x000AFFFF for video memory writes
		relative_y = y & 63;	// same as y % 64
		dest = (DWORD*) (video_addr + (relative_y << 10) + x1); // same as relative_y * 1024

		// and our offset into the picture we're working with
		srce = (DWORD*) (image_addr);
		// adjust our index into the picture to reflect line being drawn
		image_addr += width;

		// figure out the bank number & switchbank if required
		bank_num   = y >> 6;	// same as y / 64
		if (bank_num!=current_write_bank) grSwitchWriteBank(bank_num);

		// splat all DWORDS across
		for (x=0;x<dwords_to_copy;x++)
			{
			*dest = *srce;
			srce++;
			dest++;
			}

		// if we need to copy a WORD
		if (words_to_copy>0)
			{
			WORD* wdest = (WORD*) dest;
			WORD* wsrce = (WORD*) srce;
			*wdest = *wsrce;
			wdest++;
			wsrce++;

			// we may need to copy a BYTE as well
			if (bytes_to_copy>0)
				{
				BYTE* bdest = (BYTE*) wdest;
				BYTE* bsrce = (BYTE*) wsrce;
				*bdest = *bsrce;
				}
			}
		else
			// we don't need to copy a word
			// but we may still need to copy a BYTE
			{
			if (bytes_to_copy>0)
				{
				BYTE* bdest1 = (BYTE*) dest;
				BYTE* bsrce1 = (BYTE*) srce;
				*bdest1 = *bsrce1;
				}
			}
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grOverlayImage()
//
//	DESC:		Displays image on screen, ignoring all grBLACK pixels. Main use
//				is for displaying sprites... (mouse cursors??)
//
//	PARAMS:		x1,y1,x2,y2	- the rectangular region to place the image in.
//				image 		- a pointer to a grImage
//
//	NOTES:		This routine expects the image to be the same size or greater
//				than the size of the rectangular region specified by x1,y1,x2,y2
//				As such it clips only the x1,y1,x2,y2 bounds. Sorry, but its quicker!!!
//				(ie. If the image is 8x8 and you place it in a 16x16 region,
//				 all out of bounds bits will generally be utter crap)
//
//	USAGE:		grImage* pic = (grImage*) malloc(grImageSize(0,0,99,99));
//				grOverlayImage(0,0,99,99,pic);			// not clipped
//				grOverlayImage(200,200,249,249,pic);	// clipped
//				free((char*) pic);
//
//	--------------------------------------------------------------------------
void grOverlayImage(uint x1,uint y1,uint x2,uint y2,grImage* image)
	{
	uint x,y;
	uint relative_y;
	ulong bank_num;

	BYTE* dest;
	BYTE* srce;
	BYTE tmp_srce;

	ulong image_addr 	= (ulong) image + sizeof(grImageHeader);
	uint  image_width	= x2-x1+1;

	// blit the image

	for (y=y1;y<=y2;y++)
		{
		// calc address 0x000A0000 -> 0x000AFFFF for video memory writes
		relative_y = y & 63;	// same as y % 64
		dest = (BYTE*) (video_addr + (relative_y << 10) + x1); // same as relative_y * 1024

		// and our offset into the picture we're working with
		srce = (BYTE*) (image_addr);
		// adjust our index into the picture to reflect line being drawn
		image_addr += image_width;

		// figure out the bank number & switchbank if required
		bank_num   = y >> 6;	// same as y / 64
		if (bank_num!=current_write_bank) grSwitchWriteBank(bank_num);

		// we must use BYTES due to checking for transparency
		for (x=0;x<image_width;x++)
			{
			tmp_srce = *srce;	// quicker than if (*srce) *dest=*srce; due to video bus!!!
			if (tmp_srce) *dest = tmp_srce;

			dest++;
			srce++;
			}
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grGetImage()
//
//	DESC:		"Grabs" an image from the screen and stores it in an external
//				grImage. Possible uses -> sprites, editors, etc...
//
//	PARAMS:		x1,y1,x2,y2	- the rectangular region to copy the image from.
//				image 		- a pointer to a grImage
//
//	NOTES:		This routine expects the grImage to be declared _outside_ &
//				as such its behaviour with incorrect coordinates is undefined.
//
//	USAGE:		grImage* pic = (grImage*) malloc(grImageSize(0,0,99,99));
//				grGetImage(0,0,99,99,pic);
//				free((char*) pic);
//
//	--------------------------------------------------------------------------
void grGetImage(uint x1,uint y1,uint x2,uint y2,grImage* image)
	{
	uint x,y;
	uint relative_y;
	ulong bank_num;

	DWORD* dest;
	DWORD* srce;

	ulong image_addr	= (ulong) image + sizeof(grImageHeader);

	uint width 			= x2 - x1 + 1;			// total number of bytes to copy
	uint dwords_to_copy = width >> 2;			// dwords to copy == bytes / 4
	uint words_to_copy  = 0;//bytesleft & 2;		// if word-aligned, copy 1 word
	uint bytes_to_copy 	= 0;//bytesleft & 3;		// if byte-aligned, copy 1 byte

	switch (width&3)
		{
		case 0: break;
		case 1: bytes_to_copy = 1; break;
		case 2: words_to_copy = 1; break;
		case 3: words_to_copy = 1;
				bytes_to_copy = 1; break;
		}

	image->hdr.width = width;
	image->hdr.height = y2-y1 + 1;

	// blit the image

	for (y=y1;y<=y2;y++)
		{
		// calc address 0x000A0000 -> 0x000AFFFF for video memory writes
		relative_y = y & 63;	// same as y % 64
		srce = (DWORD*) (video_addr + (relative_y << 10) + x1); // same as relative_y * 1024

		// and our offset into the picture we're working with
		dest = (DWORD*) (image_addr);
		// adjust our index into the picture to reflect line being drawn
		image_addr += width;

		// figure out the bank number & switchbank if required
		bank_num   = y >> 6;	// same as y / 64
		if (bank_num!=current_read_bank) grSwitchReadBank(bank_num);

		// splat all DWORDS across
		for (x=0;x<dwords_to_copy;x++)
			{
			*dest = *srce;
			srce++;
			dest++;
			}

		// if we need to copy a WORD
		if (words_to_copy>0)
			{
			WORD* wdest = (WORD*) dest;
			WORD* wsrce = (WORD*) srce;
			*wdest = *wsrce;
			wdest++;
			wsrce++;

			// we may need to copy a BYTE as well
			if (bytes_to_copy>0)
				{
				BYTE* bdest = (BYTE*) wdest;
				BYTE* bsrce = (BYTE*) wsrce;
				*bdest = *bsrce;
				}
			}
		else
			// we don't need to copy a word
			// but we may still need to copy a BYTE
			{
			if (bytes_to_copy>0)
				{
				BYTE* bdest1 = (BYTE*) dest;
				BYTE* bsrce1 = (BYTE*) srce;
				*bdest1 = *bsrce1;
				}
			}
		}
	}

//	--------------------------------------------------------------------------
//
//	FUNC:	grImageSize()
//	DESC:	returns the size of the image
//
//	--------------------------------------------------------------------------
ulong grImageSize(uint x1,uint y1,uint x2,uint y2)
	{
	return (x2-x1+1) * (y2-y1+1) + sizeof(grImageHeader);
	}

//	--------------------------------------------------------------------------
//
//	FUNC:	grDrawLine()
//	DESC:	draws a line from x1,y1 to x2,y2 with specified color
//
//	--------------------------------------------------------------------------
void grDrawLine(uint x1, uint y1, uint x2, uint y2, uchar color)
	{
	}

//	--------------------------------------------------------------------------
//
//	FUNC:	grDrawVLine()
//	DESC:	draws a vertical line from x,y1 to x,y2 with specified color
//
//	--------------------------------------------------------------------------
void grDrawVLine(uint x,uint y1,uint y2,uchar color)
	{
	if (y1==y2) { grPutPixel(x,y1,color); return; }

	uint relative_y;
	uint bank_num;
	uint ystart = y1;
	uint yend = y2;
	if (ystart>yend) { ystart = y2; yend = y1; }
	uint xpos = x;

	for (ystart;ystart<=yend;ystart++)
		{
		relative_y = ystart & 63; 	// y % 64
		bank_num = ystart >> 6;		// y / 64

		if (bank_num!=current_write_bank) grSwitchWriteBank(bank_num);

		BYTE* dest = (BYTE*) (video_addr + (relative_y << 10) + xpos);
		*dest = color;
		}
	}

//	--------------------------------------------------------------------------
//
//	FUNC:	grDrawHLine()
//	DESC:	draws a horizontal line from x1,y to x2,y with specified color
//
//	--------------------------------------------------------------------------
void grDrawHLine(uint y,uint x1,uint x2,uchar color)
	{
	if (x1==x2) { grPutPixel(x1,y,color); return; }

	uint relative_y = y & 63;	// y % 64
	ulong bank_num = y >> 6;		// y / 64
	uint width = x2 - x1 + 1;
	uint dwords_to_copy = width >> 2; // n / 4
	uint words_to_copy = 0;
	uint bytes_to_copy = 0;
	uint x;
	WORD wcolor = color + (color<<8);
	DWORD dcolor = wcolor + (wcolor<<16);

	switch (width&3)
		{
		case 0: break;
		case 1: bytes_to_copy = 1; break;
		case 2: words_to_copy = 1; break;
		case 3: words_to_copy = 1;
				bytes_to_copy = 1; break;
		}

	if (bank_num!=current_write_bank) grSwitchWriteBank(bank_num);

	DWORD* dest = (DWORD*) (video_addr + (relative_y << 10) + x1);
	// splat all DWORDS across
	for (x=0;x<dwords_to_copy;x++)
		{
		*dest = dcolor;
		dest++;
		}

	// if we need to copy a WORD
	if (words_to_copy>0)
		{
		WORD* wdest = (WORD*) dest;
		*wdest = wcolor;
		// we may need to copy a BYTE as well
		if (bytes_to_copy>0)
			{
			BYTE* bdest = (BYTE*) wdest;
			*bdest = color;
			}
		}
	else
		// we don't need to copy a word
		// but we may still need to copy a BYTE
		{
		if (bytes_to_copy>0)
			{
			BYTE* bdest1 = (BYTE*) dest;
			*bdest1 = color;
			}
		}
	}

//	--------------------------------------------------------------------------
//
//	FUNC:	grDrawRect()
//	DESC:	draws a rectangle from x1,y to x2,y in specified color
//
//	--------------------------------------------------------------------------
void grDrawRect(uint x1, uint y1, uint x2, uint y2, uchar color)
	{
	grDrawVLine(x1,y1,y2,color);
	grDrawVLine(x2,y1,y2,color);
	grDrawHLine(y1,x1,x2,color);
	grDrawHLine(y2,x1,x2,color);
	}

//	--------------------------------------------------------------------------
//
//	FUNC:	grDrawFilledRect()
//	DESC:	draws a filled rectangle from x1,y to x2,y with specified color
//
//	--------------------------------------------------------------------------
void grDrawFilledRect( uint x1, uint y1, uint x2, uint y2,uchar color)
	{
	uint x,y;
	uint relative_y;
	ulong bank_num;

	WORD 	wfill = color + (color<<8);
	DWORD   dfill = wfill + (wfill<<16);

	DWORD* 	dest;

	uint width 			= x2 - x1 + 1;			// total number of bytes to copy
	uint dwords_to_copy = width >> 2;			// dwords to copy == bytes / 4
	uint words_to_copy  = 0;//bytesleft & 2;		// if word-aligned, copy 1 word
	uint bytes_to_copy 	= 0;//bytesleft & 3;		// if byte-aligned, copy 1 byte

	switch (width&3)
		{
		case 0: break;
		case 1: bytes_to_copy = 1; break;
		case 2: words_to_copy = 1; break;
		case 3: words_to_copy = 1;
				bytes_to_copy = 1; break;
		}

	// blit the image

	for (y=y1;y<=y2;y++)
		{
		// calc address 0x000A0000 -> 0x000AFFFF for video memory writes
		relative_y = y & 63;	// same as y % 64
		dest = (DWORD*) (video_addr + (relative_y << 10) + x1); // same as relative_y * 1024

		// figure out the bank number & switchbank if required
		bank_num   = y >> 6;	// same as y / 64
		if (bank_num!=current_write_bank) grSwitchWriteBank(bank_num);

		// splat all DWORDS across
		for (x=0;x<dwords_to_copy;x++)
			{
			*dest = dfill;
			dest++;
			}

		// if we need to copy a WORD
		if (words_to_copy>0)
			{
			WORD* wdest = (WORD*) dest;
			*wdest = wfill;
			wdest++;

			// we may need to copy a BYTE as well
			if (bytes_to_copy>0)
				{
				BYTE* bdest = (BYTE*) wdest;
				*bdest = color;
				}
			}
		else
			// we don't need to copy a word
			// but we may still need to copy a BYTE
			{
			if (bytes_to_copy>0)
				{
				BYTE* bdest1 = (BYTE*) dest;
				*bdest1 = color;
				}
			}
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grDrawString
//	DESC:	draws a string on the graphics screen using the current color
//			and current font
//
//	PARAMS:	x,y	- the coords to draw the string
//			fmt	- the format string (same as in printf)
//			... - any other parameters used by the format string
//
// --------------------------------------------------------------------------
void grDrawString(int x,int y,char *fmt, ...)
	{
	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string
	grImage*character;		// current character to be displayed
	int		currentx;		// store current x for each character
	int		i;				// counter

	va_start(argptr,fmt);	// figure out the argument starting address

	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	// and display the text onto the screen
	grHideMouse();

	currentx = x;
	// display each character of buf string
	for (i=0;buf[i]!='\0';i++)
		{
		// *NOTE:: the font is stored in the same order as the ascii set
		// but the first character is the space key 32, the list is contiguous
		// after this for 95 characters. See top of file for full list
		character = (grImage*)currentfontPointer[buf[i]-32];

		// put image on the screen
		grOverlayImage(currentx,y,currentx+character->hdr.width-1,
						y+character->hdr.height-1,character);

		// increment currentx by last char's image width
		currentx += character->hdr.width;
		}
	grShowMouse();

	va_end(argptr);				// to perform a normal return from this func
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grDrawWindow
//	DESC:	displays a window on the screen using the supplied params.
//
//	PARAMS:	x1,y1,x2,y2 - the rectangular region used by the window
//			fg,bg		- the colors used in drawing the window
//			bs			- the border style used in drawing the window
//
// --------------------------------------------------------------------------
void grDrawWindow(uint x1,uint y1,uint x2,uint y2,uchar bg,uint bs)
	{
	grHideMouse();

	// assume that black is the transparent color
	if (bg!=grBLACK) grDrawFilledRect(x1,y1,x2,y2,bg);

	switch (bs)
		{
		case bs_OUTLINE:	//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_OUTLINE,1);
							break;
		case bs_RAISED: 	//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_OUTSET,1);
							break;
		case bs_LOWERED:	//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_INSET,1);
							break;
		case bs_DBLRAISED: 	//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_OUTSET,2);
							break;
		case bs_DBLLOWERED:	//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_INSET,2);
							break;
		case bs_EDGED: 		//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_INSET,2);
							x1+=4; y1+=4; x2-=4; y2-=4;
							//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_OUTSET,2);
							break;
		case bs_FRAMED:   	//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_OUTSET,2);
							x1+=4; y1+=4; x2-=4; y2-=4;
							//MGL_drawBorderCoord(x1,y1,x2,y2,BDR_INSET,2);
		case bs_NONE:		break;
		}

	grShowMouse();
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grDrawBar
//	DESC:	displays a 3D bar on the active page.
//
//	PARAMS:	x1,y1,x2,y2 - the rectangular region to draw the bar in
//			fg,bg		- the colors used in drawing the bar
//			bs			- the border style used in drawing the bar
//
// --------------------------------------------------------------------------
void grDrawBar(uint x1,uint y1,uint x2,uint y2,uint bs)
	{
	bool singlepoint = ((x1==x2)&&(y1==y2));
	int xinc = 1;
	int yinc = 1;
	if (y1==y2) yinc=1; else yinc = 0;
	if (x1==x2) xinc=1; else xinc = 0;

/*	grHideMouse();
	switch (bs)
		{
		case bs_OUTLINE:   	//MGL_setColor(DARKCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							//MGL_setColor(LIGHTCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							break;
		case bs_RAISED :	//MGL_setColor(LIGHTCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							break;
		case bs_LOWERED:	//MGL_setColor(DARKCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							break;
		case bs_DBLRAISED:	//MGL_setColor(LIGHTCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							break;
		case bs_DBLLOWERED:	//MGL_setColor(DARKCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							break;
		case bs_EDGED:   	//MGL_setColor(DARKCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							//MGL_setColor(LIGHTCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							break;
		case bs_FRAMED:   	//MGL_setColor(LIGHTCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							//MGL_setColor(DARKCOLOR);
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
							x1+=xinc;x2+=xinc;y1+=yinc;y2+=yinc;
							if (singlepoint) //MGL_pixelCoord(x1,y1);
							else //MGL_lineCoord(x1,y1,x2,y2);
		case bs_NONE: 		break;
		}

*/	grShowMouse();
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grDrawVBar
//	DESC:	displays a vertical bar on the screen using the supplied params.
//
//	PARAMS:	xx,y1,y2 	- the coordinates used to draw the bar
//			fg,bg		- the colors used in drawing the bar
//			bs			- the border style used in drawing the bar
//
// --------------------------------------------------------------------------
void grDrawVBar(uint xx,uint y1,uint y2,uint bs)
	{
	if (y2>y1)
		grDrawBar(xx,y1,xx,y2,bs);
	else
		if (y1>y2)
			grDrawBar(xx,y2,xx,y1,bs);
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grDrawHBar
//	DESC:	displays a horizontal bar on the screen using the supplied params.
//
//	PARAMS:	yy,x1,x2 	- the coordinates used to draw the bar
//			fg,bg		- the colors used in drawing the bar
//			bs			- the border style used in drawing the bar
//
// --------------------------------------------------------------------------
void grDrawHBar(uint yy,uint x1,uint x2,uint bs)
	{
	if (x2>x1)
		grDrawBar(x1,yy,x2,yy,bs);
	else
		if (x1>x2)
			grDrawBar(x2,yy,x1,yy,bs);
	}

//	=========================================================================
//
//	FONT FUNCTIONS
//
//	=========================================================================
//
//	FUNC:		grInitFont
//	DESC:		load a font into memory and return a tag value
//
//	PARAMS:		fontname = path and name of the font in DOS
//
// --------------------------------------------------------------------------
int	grInitFont(char* fontname)
	{
	// check if we have available space in font list
	if (!(lastfont < MAX_LOADABLE_FONTS))
		return font_OutOfTags;

	lastfont++;
	// load all resources from fontname into fontpointer
	if (!(fontResource[lastfont] = new FontResourceFile(fontname,resource_READONLY,256)))
		return font_OutOfMemory;

	fontPointers[lastfont] = fontResource[lastfont]->LoadAllResources();

	// return font number
	return lastfont;
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grUseFont
//	DESC:       to set the current font to one of the "tags" returned from
//				grInitFont
//
//	PARAMS:		whichFont = tag for font
//
// --------------------------------------------------------------------------
int	grUseFont(int whichFont)
	{
	if (whichFont == font_OutOfMemory) whichFont = lastfont;

	if (whichFont > lastfont) return font_OutOfBounds;
	if (whichFont < 0) return font_OutOfBounds;

	// set fontpointer to fontpointer[whichfont]
	currentfontPointer = fontPointers[whichFont];

	return true;
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grExitFont
//	DESC:		delete the last installed font from memory
//
// --------------------------------------------------------------------------
int	grExitFont()
	{
	// if exiting current font then set current font == NULL
	if (currentfontPointer == fontPointers[lastfont])
		currentfontPointer = NULL;

	// free all resources of last loaded font
	fontResource[lastfont]->FreeAllResources((char**)fontPointers[lastfont]);
	// close font resource file
	delete fontResource[lastfont];

	// decrement number of fonts and return it
	return --lastfont;
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grExitALLFonts
//	DESC:		delete all installed fonts from memory
//
// --------------------------------------------------------------------------
int	grExitALLFonts()
	{
	// free all resources of all fonts until NoFontsLoaded
	while (grExitFont() != font_NoFontsLoaded);

	return lastfont;
	}

//	=========================================================================
//
//	SPRITE FUNCTIONS
//
//	=========================================================================
//
//	FUNC:		grCreateSprite
//	DESC:		malloc all memory associated with a sprite and
//				initialise data
//
//	PARAMS:		number_frames - number of images to be stored as animation
//				width, height - width & height of each image
//
//	NOTE:
//			For an ANIMATION sequence a SPRITE after creation should
//			have a:
//					// to give it something for first time in loop
//					grSaveUnderSprite(spr);
//
//					while(!(spr->current_frame != spr->hdr.number_frames))
//						grEraseSprite(spr);
//						grMoveSprite(spr,+4,-8);
//						spr->current_frame++;
//						grSaveUnderSprite(spr);
//						grDrawSprite(spr);
//
//	-------------------------------------------------------------------------
grSprite* grCreateSprite(char number_frames, int width, int height)
	{
	// malloc sprite
	grSprite*	dest = (grSprite*) malloc(sizeof(grSprite) + 1);

	// malloc sprite->background
	dest->background = (grImage*) malloc(grImage_HEADER_SIZE + (width * height) + 1);

	// malloc sprite->frame pointer list
	dest->frame = (grImage**) malloc(sizeof(grImage*)*number_frames + 1);

	// malloc each sprite->frame
	for (int i=0; i<number_frames; i++)
		dest->frame[i] = (grImage*) malloc(grImage_HEADER_SIZE + (width * height) + 1);

	// setup header
	dest->hdr.number_frames = number_frames;
	dest->hdr.width = width;
	dest->hdr.height = height;

	// initialise other data
	dest->x = 0;
	dest->y = 0;
	dest->oldx = 0;
	dest->oldy = 0;
	dest->current_frame = 0;
	dest->state = 0;

	// return dest
	return dest;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grDestroySprite
//	DESC:		free all memory associated with a sprite
//
//	-------------------------------------------------------------------------
void grDestroySprite(grSprite* spr)
	{
	// free each sprite->frame
	for (int i=0; i<spr->hdr.number_frames; i++)
		free(spr->frame[i]);

	// free sprite->frame pointer list
	free(spr->frame);

	// free sprite->background
	free(spr->background);

	// free sprite
	free(spr);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grEraseSprite
//	DESC:		will grPutImage the background over the sprite's position
//
//	-------------------------------------------------------------------------
void grEraseSprite(grSprite* spr)
	{
	int		x1 = spr->x,
			y1 = spr->y,
			x2 = x1 + spr->hdr.width,
			y2 = y1 + spr->hdr.height;

	// put background in old position
	grPutImage( x1,y1, x2,y2, spr->background );
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grMoveSprite
//	DESC:		saves sprite's old position then increases x & y by dx & dy
//
//	-------------------------------------------------------------------------
void	grMoveSprite( grSprite* spr, int dx, int dy )
	{
	spr->oldx = spr->x;
	spr->oldy = spr->y;

	spr->x += dx;
	spr->y += dy;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grBehindSprite
//	DESC:		will save background behind the sprite's position
//
//	-------------------------------------------------------------------------
void grSaveUnderSprite(grSprite* spr)
	{
	int		x1 = spr->x,
			y1 = spr->y,
			x2 = x1 + spr->hdr.width,
			y2 = y1 + spr->hdr.height;

	// get background
	grGetImage( x1,y1, x2,y2, spr->background );
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grDrawSprite
//	DESC:		should grPutOverlay image of sprite
//				but right now just grPutImage of a merged image
//
//	-------------------------------------------------------------------------
void grDrawSprite(grSprite* spr)
	{
//	if (spr->current_frame >= spr->hdr.number_frames) return;
	int		x1 = spr->x,
			y1 = spr->y,
			x2 = x1 + spr->hdr.width,
			y2 = y1 + spr->hdr.height;

	grOverlayImage( x1,y1, x2,y2, spr->frame[ spr->current_frame ] );
	}

// =========================================================================
//
//	PALETTE FUNCTIONS
//
// =========================================================================
//
// set palette entry #num to r,g,b
//
// --------------------------------------------------------------------------

grPalette* grCreatePalette(uint numentries)
	{
	if (numentries>256) numentries = 256;
	grPalette* p = (grPalette*) malloc(grPaletteSize(numentries));
	return p;
	}

void grDestroyPalette(grPalette* pal)
	{
	if (pal!=NULL) { free(pal); pal=NULL; }
	}

uint grPaletteSize(uint numentries)
	{
	if (numentries>256) numentries = 256;
	return numentries * 3 + 3 + sizeof(grPaletteHeader) + 1000;
	}

void grSetPaletteEntry(uchar entry,uchar r,uchar g,uchar b)
	{
	// set the palette mask register (why?? dunno)
	outp(PAL_MASK,0xFF);

	// tell card to update specific register
	outp(PAL_WRITE,entry);

	// and write the R,G,B values
	outp(PAL_DATA,r & 63);
	outp(PAL_DATA,g & 63);
	outp(PAL_DATA,b & 63);
	}

void grGetPaletteEntry(uchar entry,uchar* r,uchar* g,uchar* b)
	{
	// set the palette mask register
	outp(PAL_MASK,0xFF);

	// tell card to update specific register
	outp(PAL_READ,entry);

	// and write the R,G,B values
	*r = inp(PAL_DATA);
	*g = inp(PAL_DATA);
	*b = inp(PAL_DATA);
	}

void grFadeOutPalette(uchar start,uchar end,uchar timesin)
	{
	uchar r,g,b,decr;
	uchar* p;
	uint  index;
	float f_iter = 64 - timesin;

	// fade out the palette
	for (uint i=start;i<=end;i++)
		{
		grGetPaletteEntry(i,&r,&g,&b);

		// if (timesin==0) we save the palette ready for fade-in
		if (timesin==0)
			{

			index = i * 3;
			grPALETTE->data[index]   = r & 63;
			grPALETTE->data[index+1] = g & 63;
			grPALETTE->data[index+2] = b & 63;

			red_step[i]   = (float) r / 64.0;
			green_step[i] = (float) g / 64.0;
			blue_step[i]  = (float) b / 64.0;
			}

		r = (uchar) (f_iter * red_step[i]);
		g = (uchar) (f_iter * green_step[i]);
		b = (uchar) (f_iter * blue_step[i]);

		grSetPaletteEntry(i,r,g,b);
		}
	}

void grFadeInPalette(uchar start,uchar end,uchar timesin)
	{
	uchar r,g,b,r1,g1,b1;
	uchar* p;
	uint  index;
	float f_iter = timesin;

	// fade in the palette
	for (uint i=start;i<=end;i++)
		{
		r = (uchar) (f_iter * red_step[i]);
		g = (uchar) (f_iter * green_step[i]);
		b = (uchar) (f_iter * blue_step[i]);

		grSetPaletteEntry(i,r,g,b);
		}
	}

void grRotatePalette(uchar start,uchar end)
	{
	// some vars
	uchar r,g,b,r1,g1,b1;

	// save the start entry
	grGetPaletteEntry(start,&r,&g,&b);

	// and set all based on the next
	for (uint i=start+1;i<=end;i++)
		{
		grGetPaletteEntry(i,&r1,&g1,&b1);
		grSetPaletteEntry(i-1,r1,g1,b1);
		}

	// and set the last entry to the saved one
	grSetPaletteEntry(end,r,g,b);
	}

// =========================================================================
//
//	MOUSE FUNCTIONS
//
// =========================================================================
//
//	NOTE:	No mouse functions exist as they are simple macros reflecting
//			specified mouse functions. Refer to graphics.hpp & mouse.cpp
//			for mouse routines.
//
// --------------------------------------------------------------------------

// =========================================================================
//
//	GRID FUNCTIONS
//
// =========================================================================
//
// --------------------------------------------------------------------------
void grSetGrid(int xoffset,int yoffset,int xboxes,int yboxes,int boxwidth,int boxheight,uchar color)
	{
	GR_GRID_XOFFSET = xoffset;
	GR_GRID_YOFFSET = yoffset;
	GR_GRID_XBOXES = xboxes;
	GR_GRID_YBOXES = yboxes;
	GR_GRID_BOXWIDTH = boxwidth;
	GR_GRID_BOXHEIGHT = boxheight;
	GR_GRID_COLOR = color;
	}

void grDrawGrid()
	{
	int xx;
	int yy;

	grSetColor(GR_GRID_COLOR);

	int y1 = GR_GRID_YOFFSET;
	int y2 = GR_GRID_YOFFSET + GR_GRID_YBOXES*GR_GRID_BOXHEIGHT;
	for (int x=0;x<=GR_GRID_XBOXES;x++)
		{
		xx = x*GR_GRID_BOXWIDTH+GR_GRID_XOFFSET;
		grDrawLine(xx,y1,xx,y2);
		}

	int x1 = GR_GRID_XOFFSET;
	int x2 = GR_GRID_XOFFSET + GR_GRID_XBOXES*GR_GRID_BOXWIDTH;
	for (int y=0;y<=GR_GRID_YBOXES;y++)
		{
		yy = y*GR_GRID_BOXHEIGHT+GR_GRID_YOFFSET;
		grDrawLine(x1,yy,x2,yy);
		}
	}

void grDrawGridElem(int x,int y,uchar color)
	{
	int x1 = x*GR_GRID_BOXWIDTH+GR_GRID_XOFFSET+1;
	int y1 = y*GR_GRID_BOXHEIGHT+GR_GRID_YOFFSET+1;
	int x2 = (x+1)*GR_GRID_BOXWIDTH+GR_GRID_XOFFSET;
	int y2 = (y+1)*GR_GRID_BOXHEIGHT+GR_GRID_YOFFSET;

	grSetColor(color);
	grDrawFilledRect(x1,y1,x2,y2);
	}

void grHighlightElem(int x,int y,uchar color)
	{
	grDrawGrid();

	int x1 = x*GR_GRID_BOXWIDTH+GR_GRID_XOFFSET;
	int y1 = y*GR_GRID_BOXHEIGHT+GR_GRID_YOFFSET;
	int x2 = (x+1)*GR_GRID_BOXWIDTH+GR_GRID_XOFFSET+1;
	int y2 = (y+1)*GR_GRID_BOXHEIGHT+GR_GRID_YOFFSET+1;

	grSetColor(color);
	grDrawRect(x1,y1,x2,y2);
	}

grPoint grGridBoxSelected(grPoint p)
	{
	grPoint tmp = {-1,-1};
	// check for inside the grid...
	int x2 = GR_GRID_XOFFSET + GR_GRID_BOXWIDTH*GR_GRID_XBOXES;
	int y2 = GR_GRID_YOFFSET + GR_GRID_BOXHEIGHT*GR_GRID_YBOXES;
	if ((p.x<GR_GRID_XOFFSET)||(p.y<GR_GRID_YOFFSET)||
		(p.x>x2)||(p.y>y2))
		// returning -1,-1 if not inside
		return tmp;

	// else calc & return valid coords...
	tmp.x = (p.x-(GR_GRID_XOFFSET)) / GR_GRID_BOXWIDTH;
	tmp.y = (p.y-(GR_GRID_YOFFSET)) / GR_GRID_BOXHEIGHT;

	if (tmp.x>GR_GRID_XBOXES-1) tmp.x = GR_GRID_XBOXES -1;
	if (tmp.y>GR_GRID_YBOXES-1) tmp.y = GR_GRID_YBOXES -1;
	return tmp;
	}

// =========================================================================
//
//	DEBUGGING FUNCTIONS
//
// =========================================================================
//
//	FUNC:	DBG()
//	DESC:	clears a rectangle & draws a string on the graphics screen
//			int LIGHTRED
//
//	PARAMS:	y	- the y coord to draw the string
//			fmt	- the format string (same as in printf)
//			... - any other parameters used by the format string
//
// --------------------------------------------------------------------------
void DBG(int y,char *fmt, ...)
	{
	uchar color = grGetColor();
	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string

	va_start(argptr,fmt);	// figure out the argument starting address

	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	grSetColor(grBLACK);
	grDrawFilledRect(0,y,0,y);//MGL_textWidth(buf)+10,y+10);
	grSetColor(grLIGHTRED);

	grHideMouse();
	//MGL_moveToCoord(4,y);		// and display the text onto the screen
	//MGL_drawStr(buf);
	grShowMouse();

	va_end(argptr);				// to perform a normal return from this func

	grSetColor(color);

//    delay(500);
	}

/*void gotoxy( uchar x, uchar y )
	{
	printf("\n");
//	REGS regs;
//	memset(&regs,0,sizeof(regs));
//	regs.w.ax = 0x0200;	// function 2 - set cursor pos
//	regs.h.bh = 0;	// page #
//	regs.h.dl = (uchar) x;
//	regs.h.dh = (uchar) y;
//	int386( 0x10, &regs, &regs );
	}



void clrscr()
	{
	REGS regs;

	regs.w.ax = 0x0003;
	int386( 0x10, &regs, &regs);
	}
*/

//	-------------------------------------------------------------------------
//
//	FUNC:		grMergeImage
//	DESC:		will merge 2 images overlaying over grImage on under grImage
//
//	PARAMS:		grImage* over, under - 2 images to be merged
//	RETURNS:	grImage* dest - pointer to some malloced memory
//					NULL = error
//
//	-------------------------------------------------------------------------
grImage*	grMergeImage(grImage* over, grImage* under)
	{
	// make sure the images are the same size
	if (over->hdr.width != under->hdr.width)
		return NULL;
	if (over->hdr.height!= under->hdr.height)
		return NULL;

	// malloc the dest grImage to be returned; return NULL on fail
	grImage* dest = (grImage*)malloc(grImageSize(0,0,over->hdr.width,over->hdr.height));
	if (!dest) return dest;

	// initialise the destination bitmap grImage header
	dest->hdr.height = over->hdr.height;
	dest->hdr.width= over->hdr.width;

	// set up pointers to step through the data after the header
	char* d = (char*)&dest[grImage_HEADER_SIZE];
	char* o = (char*)&over[grImage_HEADER_SIZE];
	char* u = (char*)&under[grImage_HEADER_SIZE];

	ulong	pixelcount = 0;
	ulong	maxpixelcount = (over->hdr.width) * (over->hdr.height);

	while (pixelcount < maxpixelcount)
		{
		// set background data
		*d = *u;
		// if overlay data is not mask 0 then overide data
		if (*o)
			*d = *o;
		// increase all grImage pointers to next pixel
		*d++;
		*o++;
		*u++;

		pixelcount++;
		}
	return dest;
	}


//	-------------------------------------------------------------------
//	-------------------------------------------------------------------
//	--- New stuff for GRAPHICS.CPP ------------------------------------
//	-------------------------------------------------------------------
//	-------------------------------------------------------------------

// declarations changed from "int" to "uint"
// --------------------------------------------------------------------------
//
//	FUNC:
//	DESC:
//
// --------------------------------------------------------------------------
uint	grFontHeight()
	{
	grImage*	character = (grImage*)currentfontPointer[0];
	return character->hdr.height;
	}

// --------------------------------------------------------------------------
//
//	FUNC:
//	DESC:
//
// --------------------------------------------------------------------------
uint	grFontCharWidth(char ch)
	{
	grImage*	character = (grImage*)currentfontPointer[ch-32];
	return character->hdr.width;
	}

// --------------------------------------------------------------------------
//
//	FUNC:
//	DESC:
//
// --------------------------------------------------------------------------
uint	grFontWidth(char* st)
	{
	uint i, width = 0;

	for (i=0;st[i]!='\0';i++)
		{
		width += grFontCharWidth(st[i]);
		}

	return width;
	}

void grSetPalette(grPalette* pal,uchar start,uchar end)
	{
	uint index = 0;
	for (int i=start;i<=end;i++)
		{
		grSetPaletteEntry(i,pal->data[index],pal->data[index+1],pal->data[index+2]);
		index += 3;
		}
	}

// gets palette entry #num
void grGetPalette(grPalette* pal,uchar start,uchar end)
	{
	uint index = 0;
	for (int i=start;i<=end;i++)
		{
		grGetPaletteEntry(i,(uchar*) &pal->data[index],(uchar*) &pal->data[index+1],(uchar*) &pal->data[index+2]);
		index += 3;
		}
	}

