//	-------------------------------------------------------------------------
//
//	HANDLER.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for the Event Handler object
//					used by the EVENT MANAGER. Note that some routines
//					directly call event manager functions!
//
//	-------------------------------------------------------------------------

#include "handler.hpp"
#include "misc.hpp"
#include "eventmgr.hpp"
#include "keyboard.hpp"
#include "mouse.hpp"

#include "malloc.h"


//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for EventHandler
//	DESC:		The default constructor which defines the behaviour for
//				all derived EventHandler()'s
//
//	PARAMS:		x1,y1,x2,y2			- the coordinates of the eventhandler
//									  (required for dialog_boxes & onscreen
//				fg,bg				- the foreground & background colors
//									  used when active, swapped when inactive
//				active_border,
//				inactive_border		- the borders used when active/inactive
//				handler_num			- the handler number. THIS MUST BE THE
//									  EXACT ORDER OF CONSTRUCTION THROUGH
//									  CALLS TO EVENTMGR::ADDEVENTHANDLER()!!!
//              options				- the type of the event handler & any
//									  required options
//				zfact				- the zfactor of the handler. When
//									  interpreting mouse & kbd events for
//									  overlapping handlers, the highest
//									  zfactor is selected to handle the input
//				retevent			- the event to return on destruction
//									  NOTE: NOT YET IMPLEMENTED
//
//	-------------------------------------------------------------------------
EventHandler::EventHandler(int x1,int y1,int x2,int y2,uchar fg,uchar bg,uint active_border,uint inactive_border,
							uint handler_num,uint options,uint zfact,uint retevent)
	{
	// set our colors, borders, coordinates
	fgcolor = fg; bgcolor = bg;
	active_bs = active_border; inactive_bs = inactive_border;
	r.left = x1; r.top = y1; r.right = x2; r.bottom = y2;

	// calc our inset rectangle
	int size = max(grSizeOfBorder[active_bs],grSizeOfBorder[inactive_bs]);
	inset_r.left  = x1 + size;  inset_r.top    = y1 + size;
	inset_r.right = x2 - size; 	inset_r.bottom = y2 - size;

	// set our options, return command event & savedImage
	selected_options = options;
	returnEvent = retevent;
	savedImage = NULL;

	// set the accessorySelected to the first one
	accessorySelected = 0;

	// and set the number of accessories to 0
	numAccessories = 0;
	accessoriesAtMax = false;

	// set our handler number
	handler_number = handler_num;

	// set our zfactor
	zfactor = zfact;

	// clear our handler types
	isDialogBox = false;
	isOnScreen  = false;
	isBackgroundTask = false;
	isResponding = false;

	// if its a dialog box
	if (selected_options&evh_DIALOG_BOX)
		{
		// its a dialog box
		isDialogBox = true;

		// with these default options
		selected_options |= evh_BACKGROUND_TASK | evh_NOT_RESPONDING;
		}
	else
		// if its a background task
		if (selected_options&evh_BACKGROUND_TASK)
			{
			// its a background task
			isBackgroundTask = true;

			// with these default options
			selected_options |= evh_BACKGROUND_TASK | evh_NOT_RESPONDING;
			}
		else
			{
			// it's an onscreen task
			isOnScreen = true;

			// by default the object responds to all events
			isResponding = true;
			}

	// check whether or not we are responding
	if (selected_options&evh_NOT_RESPONDING) isResponding = false;

	// by default we save underneath the object now
	if (selected_options&evh_SAVE_UNDERNEATH) SaveUnderneath();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DESTRUCTOR for EventHandler
//	DESC:		if we saved underneath, restore it now!
//
//	-------------------------------------------------------------------------
EventHandler::~EventHandler()
	{
	// make sure we've deleted all accessories
	DelALLAccessories();

	// by default we have saved, so restore now!
	if (selected_options&evh_SAVE_UNDERNEATH) RestoreUnderneath( true );

	// return an event if required!!
//	if (returnEvent!=ev_nothing)
//		{
//		CommandEvent c = { returnEvent, NULL };
//		EVENTMGR->InsEvent(evhno_Interface,ev_command,&c);
//		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		SaveUnderneath()
//	DESC:		Saves the image of what is underneath the rectanglular
//				region of the EventHandler
//
//	-------------------------------------------------------------------------
void EventHandler::SaveUnderneath()
	{
	// get size of region to save
	ulong size = grImageSize(r.left,r.top,r.right,r.bottom);

	// allocate memory now!
	if (savedImage==NULL) savedImage = (grImage*) malloc(size);

	// save the image
	if (savedImage!=NULL)
		{
		grHideMouse();
		grGetImage(r.left,r.top,r.right,r.bottom,savedImage);
		grShowMouse();
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		RestoreUnderneath()
//	DESC:		Restores the image of what was underneath the rectanglular
//				region of the EventHandler
//
//	PARAMS:		del		- true, will delete the divot completely
//						- false, will leave divot in memory
//
//	-------------------------------------------------------------------------
void EventHandler::RestoreUnderneath(bool del)
	{
	// if we have saved into extended memory
	if (savedImage!=NULL)
		{
		// now put whats in the heap onto the screen
		grHideMouse();
		grPutImage(r.left,r.top,r.right,r.bottom,savedImage);
		grShowMouse();

		// and clean up our heap & extended memory usage...
		if (del) free(savedImage);
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DrawAllAccessories()
//	DESC:		draws all accessories
//
//	-------------------------------------------------------------------------
void EventHandler::DrawAllAccessories()
	{
	// skip out if not onScreen
	if (!isOnScreen) return;

	// skip out if no accessories
	if (numAccessories==0) return;

	// tell all accessories to draw in order of creation
	for (int acc=0;acc<numAccessories;acc++)
		accessory[acc]->Draw();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		RefreshAllAccessories()
//	DESC:		Refreshes all accessories
//
//	-------------------------------------------------------------------------
void EventHandler::RefreshAllAccessories()
	{
	// skip out if not onScreen
	if (!isOnScreen) return;

	// skip out if no accessories
	if (numAccessories==0) return;

	// tell all accessories to refresh in order of creation
	for (int acc=0;acc<numAccessories;acc++)
		accessory[acc]->Refresh();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Select()
//	DESC:		This function is called by the Event Manager upon its
//				selecting a new event handler for keyboard/mouse input...
//
//	PARAMS:		which	- which accessory to select
//						  evh_FIRST_ACCESSORY,
//						  evh_LAST_ACCESSORY or
//						  evh_CURRENT_ACCESSORY
//
//	-------------------------------------------------------------------------
void EventHandler::Select( uint which )
	{
	// this is only for onscreen objects
	if (!isOnScreen) return;

	grSetColor(LIGHTRED);
	grDrawRect(r.left,r.top,r.right,r.bottom);

	if (numAccessories>0)
		{
		switch (which)
			{
			case	evh_FIRST_ACCESSORY:	accessorySelected = 0;
											break;
			case	evh_CURRENT_ACCESSORY:	break;
			case	evh_LAST_ACCESSORY:		;
			default:						accessorySelected = numAccessories-1;
											break;
			}

		accessory[accessorySelected]->Select();
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DeSelect()
//	DESC:		This function is called by the Event Manager upon its
//				selecting a new event handler for keyboard/mouse input...
//
//	-------------------------------------------------------------------------
void EventHandler::DeSelect()
	{
	// this is only for onscreen objects
	if (!isOnScreen) return;

	grSetColor(BLACK);
	grDrawRect(r.left,r.top,r.right,r.bottom);

	if (numAccessories>0)
		accessory[accessorySelected]->DeSelect();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		SelectNextAccessory()
//	DESC:		Selects the next (selectable) accessory.
//
//	-------------------------------------------------------------------------
void EventHandler::SelectNextAccessory()
	{
	// tell the previous accessory to change to deselected mode
	accessory[accessorySelected]->DeSelect();

	// use temp value for currently selected accessory
	int access = accessorySelected;

	do
		{
		// get the next accessory
		access++;

		// wrap iff at the last accessory
		if (access>=numAccessories) access=0;
		}
	// keep looping until we find a selectable accessory
	while ((!accessory[access]->isSelectable)&&(access!=accessorySelected));

	// update our selected accessory
	accessorySelected = access;

	// tell the accessory that it is selected
	accessory[accessorySelected]->Select();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		SelectPrevAccessory()
//	DESC:		Selects the previous (selectable) accessory.
//
//	-------------------------------------------------------------------------
void EventHandler::SelectPrevAccessory()
	{
	// tell the previous accessory to change to deselected mode
	accessory[accessorySelected]->DeSelect();

	// use temp value for currently selected accessory
	int access = accessorySelected;

	do
		{
		// get the previous accessory
		access--;

		// check to see if we've gone past the first accessory
		if (access<0) access = numAccessories-1;
		}
	// keep looping until we find a selectable accessory
	while ((!accessory[access]->isSelectable)&&(access!=accessorySelected));

	// update our selected accessory
	accessorySelected = access;

	// tell the accessory that it is selected
	accessory[accessorySelected]->Select();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		WhichAccessorySelected()
//	DESC:		Calculates which accessory was selected based upon a mouse
//				location point.
//
//	PARAMS:		ms_point	- the point to check
//								eg. int which = WhichAccessorySelected(e.mouse.where);
//
//	RETURNS		>=0 - the number of the accessory selected
//				 -1 - the event handler (window) was selected
//
//	-------------------------------------------------------------------------
int EventHandler::WhichAccessorySelected(grPoint &ms_point)
	{
	// if there are no accessories, return -1
	if (numAccessories==0) return -1;

	// scan through each accessory
	for (int i=0;i<numAccessories;i++)
		// checking if the point is inside the bounds of the accessory
		if ((ms_point.x>=accessory[i]->r.left)&&(ms_point.x<=accessory[i]->r.right)&&
			(ms_point.y>=accessory[i]->r.top)&&(ms_point.y<=accessory[i]->r.bottom))
			// if the point is inside, return the accessory number
			return i;

	// otherwise the point was not inside any of the accessories
	return -1;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		AddAccessory()
//	DESC:		Adds an accessory to the window to be used automatically by
//				EventHandler::HandleEvent()
//
//	PARAMS:		accessry - the result from a  new Accessory()
//						   eg. uint OK = AddAccessory( new OKButton(,,,) );
//
//	RETURNS:			- a pointer to the accessory.
//				NULL 	- cannot create accessory (too many)
//
//	-------------------------------------------------------------------------
Accessory* EventHandler::AddAccessory(Accessory* accessry)
	{
	// if we are already at max delete the new'd accessory & return
	if (accessoriesAtMax)
		{
		delete accessry;
		return NULL;
		}

	// add an accessory
	accessory[numAccessories] = accessry;

	// and increment the number of accessories
	numAccessories++;

	// check if we are at max.
	accessoriesAtMax = (numAccessories == MAX_ACCESSORIES_IN_HANDLER);

	// modify the accessory to assume GLOBAL screen coordinates!
	accessry->AdjustCoords(inset_r.left,inset_r.top);

	// and return our pointer
	return accessry;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DelAccessory()
//	DESC:		Deletes the last created accessory from the window.
//
//	-------------------------------------------------------------------------
void EventHandler::DelAccessory()
	{
	// if there are no accessories, skip out
	if (numAccessories==0) return;

	// decrement the number of accessories
	numAccessories--;

	// and delete the accessory
	delete accessory[numAccessories]; accessory[numAccessories] = NULL;

	// check if we are at max.
	accessoriesAtMax = (numAccessories == MAX_ACCESSORIES_IN_HANDLER);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DelALLAccessories()
//	DESC:		Deletes all accessories
//
//	-------------------------------------------------------------------------
void EventHandler::DelALLAccessories()
	{
	// use DelAccessory to delete all accessories
	while (numAccessories>0) DelAccessory();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Animate()
//	DESC:		This function is called upon receiving a cmdAnimate event
//				This occurs automatically by calling
//				EventHandler::HandleEvent() in derived HandleEvent()'s
//
//	-------------------------------------------------------------------------
void EventHandler::Animate()
	{
	// does nothing
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Draw()
//	DESC:		This function completely draws the EventHandler on the screen
//				by drawing the border/background/etc, then calling Refresh()
//				to update the contents of the window!
//
//	-------------------------------------------------------------------------
void EventHandler::Draw()
	{
	// skip out if not onScreen
	if (!isOnScreen) return;

	// draw the appropriate window!
	grDrawWindow(r.left,r.top,r.right,r.bottom,bgcolor,active_bs);

	DrawAllAccessories();
	// highlight the accessory currently selected
//	accessory[accessorySelected]->Select();

	Refresh();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Refresh()
//	DESC:		This function fills in the information for a window, allowing
//				almost flicker-free window updates!
//
//	-------------------------------------------------------------------------
void EventHandler::Refresh()
	{
	// skip out if not onScreen
	if (!isOnScreen) return;

	// refresh the insides of all accessories
	if (selected_options&evh_REFRESH_ACCESSORIES) RefreshAllAccessories();
	else DrawAllAccessories();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		HandleEvent()
//	DESC:		This is the generic HandleEvent() function which should be
//              called by any derived objects HandleEvent() function as follows
//
//				eg. MyWindow::HandleEvent(Event &e)
//						{
//
//						(interpret events here to override EventHandler)
//
//						// call parent HandleEvent() to handle accessories!!
//						EventHandler::HandleEvent(e);
//
//						(interpret events here after accessories handle event)
//
//						}
//
//				by default this HandleEvent interprets the following keys:
//					TAB 	- move to next accessory
//					SH_TAB 	- move to previous accessory
//					RET 	- select current accessory
//					SPACE   - select current accessory
//
//	PARAMS:		e - the specified event to be handled
//
//	-------------------------------------------------------------------------
void EventHandler::HandleEvent(Event &e)
	{
	// skip out if we aren't meant to be here....
	if (!isResponding)
		if (e.what!=ev_broadcast) return;

	// handle the event!
	int acc;
	switch (e.what)
		{
		// handle mouse events
		case ev_mouse:		// find selected accessory
							acc = WhichAccessorySelected(e.mouse.where);

							// if an accessory was selected
							if (acc>=0)
								{
								// and it isn't the currently selected accessory
								if (acc!=accessorySelected)
									{
									// deselect the current accessory
									accessory[accessorySelected]->DeSelect();

									// select the accessory if left button down
									if (e.mouse.event==mb_leftdn)
										accessory[acc]->Select();
									}

								// make sure we have selected the accessory
								if (!accessory[acc]->isSelected)
									accessory[acc]->Select();

								// and tell it to handle the event
								accessorySelected = acc;
								accessory[acc]->HandleEvent(e);
								}
							else
								// no accessory was selected
								// so if there are any accessories, deselect the current one
								if (numAccessories>0)
									if (accessory[accessorySelected]->isSelected)
										accessory[accessorySelected]->DeSelect();

							break;

		// handle keyboard events
		case ev_keyboard:	switch (e.keyboard.event)
								{
								// make sure the appropriate accessory is selected
								case	TAB_UP:	if (numAccessories>0)
													if (!accessory[accessorySelected]->isSelected)
														accessory[accessorySelected]->Select();
												break;
								// see if the accessory can handle the event
								default:   		if (numAccessories>0) accessory[accessorySelected]->HandleEvent(e);
												break;
								}
							break;

		// handle command events
		case ev_command:  	switch (e.command.event)
								{
								case cmdAnimate:	Animate(); break;
								case cmdRefresh:	Refresh(); break;
								}
							break;

		// handle broadcast events
		case ev_broadcast:  switch (e.broadcast.event)
								{
								case cmdAnimate:	Animate(); break;
								case cmdRefresh:	Refresh(); break;
								}
							break;

		// handle timer events
		case ev_timer:  	switch (e.timer.event)
								{
								case cmdAnimate:	Animate(); break;
								case cmdRefresh:	Refresh(); break;
								}
							break;
		}
	}
