//	-------------------------------------------------------------------------
//
//	HANDLER.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains defines for the Event Handler object
//					used by the EVENT MANAGER.  Note that some routines
//					directly call event manager functions!
//
//	-------------------------------------------------------------------------

#ifndef __handler_hpp
#define __handler_hpp

#include "event.hpp"
#include "accessry.hpp"

#include "graphics.hpp"

// options for event handlers

#define evh_ONSCREEN_TASK			0x0001	// handler is onscreen (default)
#define	evh_BACKGROUND_TASK			0x0002	// handler is background task
#define evh_DIALOG_BOX				0x0004	// handler is dialog box

#define evh_NOT_RESPONDING			0x0008  // does not respond to mouse/kbd
#define evh_SAVE_UNDERNEATH			0x0010	// save underneath upon creation
#define evh_REFRESH_ACCESSORIES		0x0020	// refresh() accessories only

// the maximum number of accessories in any specific event handler

#define MAX_ACCESSORIES_IN_HANDLER		64

// options for Select() (used internally)

#define evh_FIRST_ACCESSORY			0		// select
#define evh_LAST_ACCESSORY          1
#define	evh_CURRENT_ACCESSORY		2

//	-------------------------------------------------------------------------
//
//	CLASS:	EventHandler
//
//	DESC:   This is the base object which can respond to events. The event
//			handler can be of the type:
//
//				onscreen			- handler is displayed on the screen
//									  and responds to all event types.
//				background_task		- handler is "offscreen" and does not
//									  interpret mouse & keyboard input.
//				dialog_box			- this type of handler is basically a
//									  background task that waits for an
//									  "opendialogbox" command and then
//									  becomes an onscreen task with exclusive
//									  access to mouse & kbd. Upon receiving
//									  a "closedialogbox" command it returns
//									  to a background task.
//
//			All eventhandlers may contain "accessories" which can be any
//			type of input device responding to mouse & kbd. These can
//			range from a simple TextInputButton through to a TextScrollBox
//			through to a spaceship! :-)
//
//	-------------------------------------------------------------------------

class EventHandler
	{
	protected:
		uint	returnEvent;        // command event returned (optional)

		uchar   fgcolor;			// the foreground color
		uchar	bgcolor;			// the background color
		uint	active_bs;			// the active border style
		uint	inactive_bs;		// the inactive border style

		grImage* savedImage;		// whats underneath the window!
		uint	selected_options;	// the options selected for the EventHandler

		// storage of accessories
		Accessory*	accessory[MAX_ACCESSORIES_IN_HANDLER];


		void 	DrawAllAccessories();	// draws all accessories
		void	RefreshAllAccessories();// refreshes all accessories

		// which accessory was selected by the user
		int		WhichAccessorySelected(grPoint &mouse_pos);

		void 	SaveUnderneath();		// save what is underneath the window
		void 	RestoreUnderneath( bool del );	// restore underneath the window

	public:
	// NOTE:	the following are used (internally) by EVENTMGR
		uint 	zfactor;			// the z dimension (x,y,z)

		uint	handler_number;		// the handler number

		int		numAccessories;		// number of accessories contained in hanlder
		int		accessorySelected;  // accessory which is currently selected for
									// keyboard response

		void 	SelectNextAccessory();	// select the next (selectable) accessory
		void	SelectPrevAccessory();	// select the prev (selectable) accessory
	// END NOTE:

		grRect	r;				// rectangular region taken up by handler
		grRect	inset_r;		// rectangular region inside borders, etc.

		bool	isResponding;   // true -> responds to events
								// false -> ignore ALL events except ev_broadcast

		bool	isOnScreen;		// true -> object is an on screen handler
								// false -> object is background task
		bool	isBackgroundTask;	// true -> object is a background task
									// false -> object is onscreen
		bool	isDialogBox;	// true -> event handler is a dialog box
								// false -> other

		bool	accessoriesAtMax;	// true if no more accessories can be added

		EventHandler(int x1,int y1,int x2,int y2,uchar fg,uchar bg,
					 uint active_border,uint inactive_border,
					 uint handler_num,uint options,uint zfact,uint retevent);
		virtual ~EventHandler();

		virtual void Select( uint which );		// select the event handler
		virtual void DeSelect();				// deselect the event handler

		Accessory* AddAccessory(Accessory* accessry); // add an accessory to the handler
		void 	DelAccessory();					// deletes the last accessory
		void	DelALLAccessories();			// deletes all accessories

		virtual void Animate();  				// animates (automatically)
		virtual void Draw();					// draw for first time in
		virtual void Refresh();					// draw for other times in
		virtual void HandleEvent(Event& e);		// handles the current event
	};


#endif

