//	-------------------------------------------------------------------------
//
//	KEYBOARD.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for a keyboard handler (ISR)
//					which generates events into the EVENTQUEUE
//
//	-------------------------------------------------------------------------

#include "keyboard.hpp"
#include "mouse.hpp"
#include "event.hpp"
#include "stddefns.hpp"
#include "eventmgr.hpp"

#include "stdlib.h"
#include "conio.h"
#include "stdio.h"
#include "mem.h"
#include "dos.h"

// some interrupt & port numbers

#define	KEYBOARD_INT			0x09	// keyboard interrupt no.
#define	KEY_BUFFER				0x60	// location of keyboard buffer
#define	KEY_CONTROL				0x61	// location of keyboard controller
#define INT_CONTROL				0x20	// location of interrupt controller

// storage of previous keyboard ISR (Interrupt Service Routine)
void _interrupt (_far *Old_Keyboard_Isr)();

// a KeyboardEvent variable used when generating new events
KeyboardEvent	keyboard_EventInfo;

bool Quit = false;

// GLOBAL storage of keypress information
uint	raw_key;

// key_table special identifiers

#define FIRST_KEY_INDEX		96
#define SECOND_KEY_INDEX	42

// key table extended keys

#define	LEFT_SHIFT			100
#define RIGHT_SHIFT			101
#define LEFT_CTRL			102
#define RIGHT_CTRL			103
#define LEFT_ALT			104
#define RIGHT_ALT			105	// last key in key_table!
#define ENTER				28
#define SLASH				53

// modified key flag used when first_key_index is found
bool	modified_key;

// the status of a key_table index
#define key_neither		0		// special use for non-auto-repeating
#define key_released	1		// key is released
#define key_pressed		2		// key is pressed
//$AUTOuint	key_auto		= 40;	// key is auto repeating

// storage for the current state of each key
uint	key_table[106];
uint 	table_index = 0;		// and a GLOBAL index used in the ISR

// a flag used to test whether or not the handler is installed!
bool	keyboard_HANDLER_INSTALLED = false;

// the key-to-ASCII table.
const uchar	ascii_table[] =
	{ 0,27,'1','2','3','4','5','6','7','8','9','0','-','=',8,9,
	  'q','w','e','r','t','y','u','i','o','p','[',']',13,0,
	  'a','s','d','f','g','h','j','k','l',';',39,'`',0,92,
	  'z','x','c','v','b','n','m',',','.','/' };
const uchar	shift_ascii_table[] =
	{ 0,27,'!','@','#','$','%','^','&','*','(',')','_','+',0,0,
	  'Q','W','E','R','T','Y','U','I','O','P','{','}',13,0,
	  'A','S','D','F','G','H','J','K','L',':','"','~',0,'|',
	  'Z','X','C','V','B','N','M','<','>','?' };

// a flag used to check for the first ENTER keypress [as we install our
// keyboard handler so quickly, generally the BIOS has not yet interpreted
// the ENTER_UP function!!! As such we must ensure that we ignore the first
// instances of ENTER_UP]
bool 	keyboard_FIRST_TIME_ENTER = true;

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_IsReleased()
//	DESC:		returns true if the specified keypress is a key_released event
//
//	-------------------------------------------------------------------------
bool keyboard_IsReleased( uint keyvalue )
	{
	return (keyvalue>=1000);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_IsPressed()
//	DESC:		returns true if the specified keypress is a key_pressed event
//
//	-------------------------------------------------------------------------
bool keyboard_IsPressed( uint keyvalue )
	{
	return (keyvalue<1000);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_ASCII()
//	DESC:		returns the ascii value of a keyboard return value. Also
//				takes into account the status of the SHIFT keys.
//				eg. keyboard_ASCII( RET_DN, keyboard_SHIFT ) == 13
//					keyboard_ASCII( A_UP, keyboard_SHIFT ) == 'A'
//
//	-------------------------------------------------------------------------
uchar keyboard_ASCII( uint keyvalue, uint modifiers )
	{
	int keyval = keyvalue;

	// convert from a "_UP" value to a "_DN" value
	if (keyval>1000) keyval -= 1000;

	// handle special case for space bar
	if (keyval==SPACE_DN) return 32;

	// skip out if we shouldn't check the keyboard value
	if (keyval>53) return 0;

	// quickly compare and return any shift-modified key values
	if (modifiers & keyboard_SHIFT)
		keyval = shift_ascii_table[keyval];
	else
		// return any normal key values
		keyval = ascii_table[keyval];

	// if we have found an incorrect translation, set to original input!
	if (keyval==0) keyval = keyvalue;

	// return the converted value
	return keyval;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_GetState()
//	DESC:		returns the current state of the shift/alt/ctrl keys
//
//	-------------------------------------------------------------------------
uint keyboard_GetState()
	{
	// assume no modifiers
	uint modifiers =  0x0000;

	// scan the appropriate key_table indexes and reflect whats down
	if (key_table[LEFT_SHIFT]==key_pressed) 	modifiers |= keyboard_LEFT_SHIFT;
	if (key_table[RIGHT_SHIFT]==key_pressed)    modifiers |= keyboard_RIGHT_SHIFT;
	if (key_table[LEFT_ALT]==key_pressed)       modifiers |= keyboard_LEFT_ALT;
	if (key_table[RIGHT_ALT]==key_pressed)      modifiers |= keyboard_RIGHT_ALT;
	if (key_table[LEFT_CTRL]==key_pressed)      modifiers |= keyboard_LEFT_CTRL;
	if (key_table[RIGHT_CTRL]==key_pressed)     modifiers |= keyboard_RIGHT_CTRL;

	// and return our modifiers flag
	return modifiers;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_SetAutoRepeat()
//	DESC:		sets the delay for the auto repeat on key_dn
//
//	-------------------------------------------------------------------------
void keyboard_SetAutoRepeat( uint rate )
	{
//$AUTO	key_auto = rate;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_GetEvent()
//	DESC:		checks to see if any keys have been pressed & generates
//				an appropriate keyboard event into the EVENT QUEUE
//
//	-------------------------------------------------------------------------
void keyboard_GetEvent()
	{
	// get the keyboard modifiers
	keyboard_EventInfo.modifiers = keyboard_GetState();

	// get where the mouse is
	keyboard_EventInfo.where = mouse_GetPos();

	// scan through our key table and check each key to see if it is up or
	// down. For simplicity, all keys auto-repeat if held down.

	int index = 0;

	do
		{
		// if the key is not pressed or released!
		if (key_table[index]!=key_neither)
			{
			// check for the very first time ENTER is released (refer above)!
			if ((keyboard_FIRST_TIME_ENTER)&&(index==ENTER))
				{
				key_table[ENTER] = key_neither;
				keyboard_FIRST_TIME_ENTER = false;
				continue;
				}

			// get the next available keypress from the buffer
			// and if its pressed, leave it pressed and generate an event
//$AUTO			if (key_table[index]>=key_pressed)
			if (key_table[index]==key_pressed)
				{
				keyboard_EventInfo.event = index;

				// increment our auto repeat value
//$AUTO				key_table[index]++;
//$AUTO				if (key_table[index]>=key_auto)
					// auto repeat, simply generate a key_up value &
					// reset the keypress info. to key_pressed.
//$AUTO					{
//$AUTO					keyboard_EventInfo.event = index + 1000;
//$AUTO					key_table[index] = key_pressed;
//$AUTO					}
				}
			else
				// since it has been released, we set the table entry to
				// key_neither, thus we no longer generate any more
				// _key_is_released_ events!!!
				{
				keyboard_EventInfo.event = index + 1000;
				key_table[index] = key_neither;
				}

			// if we don't already have this event in the queue
//			printf("\nKEYPRESS == %d",keyboard_EventInfo.event);
//			if (keyboard_EventInfo.event==ESC_DN) Quit = true;
			if (!EVENTMGR->EventInQueue(ev_keyboard,&keyboard_EventInfo))
				// place it onto the queue
				EVENTMGR->GenEvent(0,ev_keyboard,&keyboard_EventInfo);
			}

		// do the next iteration
		index++;
		}

	// ignore keyboard modifiers!
	while (index<100);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_ISR()
//	DESC:		this is the keyboard ISR function which interprets most of
//				the keyboard and generates into a table appropriate responses.
//
//	NOTES:		The keyboard generates a scan code, generally 1 byte, which
//				represents the key either being pressed (<128) or released
//				(>127). This value is stored in a table which we can later
//				process as needed.
//
//				Some keys require special interpretation, because the keyboard
//				sends the scan codes for some keys as 2 bytes (2 interrupts!)
//				or as 4 bytes (4 interrupts?).
//
//				The following keys generate the prefix scan code of 96:
//
//				  key     		prefix      code
//
//				RIGHT_CTRL		96 			29
//				RIGHT_ALT		96 			56
//				KEYPAD_ENTER	96 			28
//				KEYPAD_SLASH	96 			53
//
//				The following keys generate the prefix scan codes - 96 42 96
//
//				INS				96 42 96    82
//				DEL             96 42 96    83
//				HOME            96 42 96    71
//				END             96 42 96    79
//				PGUP            96 42 96    73
//				PGDN            96 42 96    81
//				UP              96 42 96    72
//				DN              96 42 96    80
//				LT              96 42 96    95
//				RT              96 42 96    77
//
//				The way we interpret the other keys is simply to interpret
//				their code...
//
//	-------------------------------------------------------------------------
void _interrupt _far keyboard_ISR()
	{
	// skip out if we have not yet finished installation!
	if (!keyboard_HANDLER_INSTALLED) return;

	// talk to the keyboard & get the keypress
	raw_key = inp(KEY_BUFFER);			// get the raw keypress
	char tmp = inp(KEY_CONTROL) | 0x82;	// read the control register &
										// set bits to reset flip-flop

	outp(KEY_CONTROL,tmp);				// reset the flip-flop
	tmp &= 0x7F;
	outp(KEY_CONTROL,tmp);

	// calc our look up index into keypress table 0-101
	table_index = (raw_key & 0x007F);	// mask out hi bit of lo byte :-)

	// figure out if the key is pressed or released
	int value = key_pressed;
	if (raw_key>127) value = key_released;

	// if scan code 96 we set a flag for next time in
	if (table_index==FIRST_KEY_INDEX) modified_key = true;
	else
		{
		// next (or first :-) ) time in... is it INS/DEL/etc? Or maybe LEFT_SHIFT
		if (table_index==SECOND_KEY_INDEX)
			{
			// if we previously found a 96 code
			if (modified_key)
				{
				// simply ignore the 96 42 portion of prefix
				modified_key = false;
				}
			else
				// it must be the left shift key (also 42)
				key_table[LEFT_SHIFT] = value;
			}
		else
			{
			// check for easier prefix 96
			if (modified_key)
				{
				// and check for what it modifies
				switch (table_index)
					{
					// setting our key_table appropriately
					case	29:		key_table[RIGHT_CTRL] = value;	break;
					case	56:		key_table[RIGHT_ALT] = value;	break;
					case	28:		key_table[ENTER] = value;		break;
					case	53:		key_table[SLASH] = value;		break;
					// note: we translate UP/DN/etc. to KP8/KP2/etc. by default
					default:		key_table[table_index] = value; break;
					}
				// get ready for the next key
				modified_key = false;
				}
			else
				// its a normal (1-byte) scan code
				{
				switch (table_index)
					{
					// check for special keys which generate a single byte now
					case	29:		key_table[LEFT_CTRL] = value;	break;
					case	56:		key_table[LEFT_ALT] = value;	break;
					case	54:		key_table[RIGHT_SHIFT] = value;	break;
					// otherwise it is some other keyboard key...
					default:		key_table[table_index] = value;
					}
				}
			}
		}

	// re-enable interrupts
	outp(INT_CONTROL,0x20);	// signal end-of-interrupt
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_InstallHandler()
//	DESC:		installs our custom keyboard handler.
//
//	-------------------------------------------------------------------------

void keyboard_InstallHandler()
	{
	if (!keyboard_HANDLER_INSTALLED)
		{
		// flush the keyboard buffer caJUSTse
		while (kbhit()) getch();

		// clear keypress table (assumes all keys released)
		memset(&key_table,key_neither,sizeof(key_table));

		// save old keyboard ISR
		Old_Keyboard_Isr = _dos_getvect(KEYBOARD_INT);

		// and replace it with ours!
		_dos_setvect(KEYBOARD_INT,keyboard_ISR);

		// make sure the keyboard handler is reset
		atexit(keyboard_RemoveHandler);

		// set a flag for check later
		keyboard_HANDLER_INSTALLED = true;

		// set the auto repeat rate
//$AUTO
//		keyboard_SetAutoRepeat( 350 );
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_RemoveHandler()
//	DESC:		Removes our custom keyboard handler & sets the handler back
//				to the original (BIOS/DOS/???) keyboard handler.
//
//	-------------------------------------------------------------------------
void keyboard_RemoveHandler()
	{
	if (keyboard_HANDLER_INSTALLED)
		{
		// restore old keyboard ISR
		_dos_setvect(KEYBOARD_INT,Old_Keyboard_Isr);

		// and clear our handler_installed flag!!!
		keyboard_HANDLER_INSTALLED = false;
		}
	}


/*

// a dummy MAIN()

void main()
	{
//	clrscr();

	keyboard_InstallHandler();

	do
		{
//		delay(100);
		keyboard_GetEvent();

//		Quit = ((raw_key==1)||(raw_key==129));
		}
	while (!Quit);

	keyboard_RemoveHandler();
	}

*/