//	-------------------------------------------------------------------------
//
//	RESOURCE.HPP
//
//	-------------------------------------------------------------------------
//
//	DESCRIPTION:	This file defines the Resource object which is used to
//					access a pseudo-random-access file containing variable
//					length records (resources)
//
//	-------------------------------------------------------------------------

#ifndef __resource_hpp
#define __resource_hpp

#include "dosfile.hpp"

#include "stdio.h"
#include "io.h"

// error messages for the file object
#define resource_NoError				dosfile_NoError
#define resource_FileNotFound			dosfile_FileNotFound
#define resource_SeekError				dosfile_SeekError
#define resource_FileIsWriteOnly		dosfile_FileIsWriteOnly
#define resource_FileIsReadOnly			dosfile_FileIsReadOnly
#define resource_ReadError				dosfile_ReadError
#define resource_WriteError				dosfile_WriteError
#define resource_FileNotOpen			dosfile_FileNotOpen
#define resource_FileAlreadyOpen		dosfile_FileAlreadyOpen
#define resource_FileAlreadyClosed		dosfile_FileAlreadyClosed

#define resource_ResourceNotFound			-11
#define resource_ResourceAlreadyUsed		-12

// options used when opening the file
#define resource_READWRITE			dosfile_READWRITE	// read-write access
#define resource_READONLY			dosfile_READONLY	// read-only access
#define resource_CREATE				dosfile_CREATE		// create new file (overwrites existing file)

// the types of resources
#define resource_FONT_RESOURCE				0
#define resource_BITMAP_RESOURCE			1
#define resource_TEXT_RESOURCE				2
#define resource_SOUND_RESOURCE				3
#define resource_PALETTE_RESOURCE			4
#define resource_GENERAL_RESOURCE			100			// can be used however!!

// this is the information for a specified record
struct ResourceIndex
	{
	long 	location; 	// the absolute start location of the record
	long	size;		// the size of the record
	};

// this is the header of a resource file
struct ResourceHeader
	{
	uint			type;			// the type of resource
	long 			max_resources;	// the maximum number of records in the resource
	long			num_resources;	// the number of records in the resource
	};

class ResourceFile : public DOSFile
	{
	protected:
		ResourceHeader		header;
		ResourceIndex		index;

		int		ReadHeader();       // read the resource header
		int		WriteHeader();		// write the resource header

		int 	ReadResourceIndex(long resource_num);	// read a specified resource's index
		int		WriteResourceIndex(long resource_num);  // write a specified resource's index

		virtual int		CompressResource() { return 0; }   // implemented in
		virtual int		DecompressResource() { return 0; }	// derived classes

	public:
		ResourceFile(DirString* pathname,uint options,long num_resources = 256,uint type = resource_BITMAP_RESOURCE); 	// defines the ResourceFile with options
		ResourceFile(const char* pathname,uint options,long num_resources = 256,uint type = resource_BITMAP_RESOURCE);	// defines the ResourceFile with options
		virtual ~ResourceFile(); // close the ResourceFile

		// returns the number of resources currently in file [append == WriteResource(NumResources()+1,...)]
		long 	NumResources();
		long	ResourceSize(long resource_num);

		virtual long	SizeOfResource(long resource_num);
		virtual int		ReadResource(long resource_num,char* buffer);
		virtual int		WriteResource(long resource_num,long size,char* buffer);

		virtual	char**	LoadAllResources();
		virtual void	FreeAllResources(char** datalist);
	};


#endif