//	--------------------------------------------------------------------------
//
//	SPRITES.CPP
//
//	--------------------------------------------------------------------------

#include	"sprites.hpp"
#include	"pcx.hpp"

SpriteResourceFile::SpriteResourceFile(DirString* pathname,uint options,long num_resources)
					: ResourceFile(pathname,options,num_resources,resource_BITMAP_RESOURCE)
	{
	}

SpriteResourceFile::SpriteResourceFile(	const char* pathname,uint options,long num_resources)
					: ResourceFile(	pathname,options,num_resources,resource_BITMAP_RESOURCE)
	{
	}

SpriteResourceFile::~SpriteResourceFile()
	{
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	SizeOfResource
//	DESC:	use to malloc the size of image in a resource
//
//	-------------------------------------------------------------------------
spriteHeader SpriteResourceFile::GetResourceHeader (long resource_num)
	{
	spriteHeader	head;
	head.number_frames = 0;
	head.width = 0;
	head.height = 0;

	// read the resource index
	if (ReadResourceIndex( resource_num ) != resource_NoError) return head;

	// if the size is zero, there is no said resource
	if (index.size==0) return head;

	// seek to the start of the resource
	if (Seek( index.location ) != resource_NoError) return head;

	if (DOSFile::Read( (char*)&head, sizeof(spriteHeader) ) != resource_NoError) return head;

	return head;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		ReadResource
//	DESC:       reads a resource of 'size' bytes, decompresses it and
//				places them in an image 'buffer'
//
//	-------------------------------------------------------------------------
int SpriteResourceFile::ReadResource(long resource_num,sprite* spr)
	{
	// skip out if we're write-only
	if (fileOptions&dosfile_WRITEONLY) return fileError = resource_FileIsWriteOnly;

	// read the resource status
	if ((fileError = ReadResourceIndex( resource_num )) != resource_NoError) return fileError;

	// if the size is zero, there is no said resource
	if (index.size==0) return fileError = resource_ResourceNotFound;

	// seek to the start of the resource
	if ((fileError = Seek( index.location )) != resource_NoError) return fileError;

//	fileError = DOSFile::Read(tempbuffer, index.size );

	// read the resource

	// read the header in first
	if ((fileError = DOSFile::Read( (char*)&spr->hdr,sizeof(spriteHeader) )) != resource_NoError) return fileError;

	// calculate the size of each frame
	long frame_size = grImage_HEADER_SIZE + ( spr->hdr.width * spr->hdr.height );

	// read each frame in
	for (int i=0; i<spr->hdr.number_frames; i++)
		{
		if ((fileError = DOSFile::Read( (char*)spr->frame[i], frame_size )) != resource_NoError) return fileError;
		}

	// return our last file error
	return fileError;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		WriteResouce()
//	DESC:       writes a resource of 'size' bytes starting at 'buffer'
//
//	-------------------------------------------------------------------------
int SpriteResourceFile::WriteResource(long resource_num, sprite* spr)
	{
	// skip out if we're read-only
	if (fileOptions&dosfile_READONLY) return fileError = resource_FileIsReadOnly;

	// read the resource status
	if ((fileError = ReadResourceIndex( resource_num )) != resource_NoError) return fileError;

	// calculate the size of writable data
	long frame_size = grImage_HEADER_SIZE + ( spr->hdr.width * spr->hdr.height );
	long size = sizeof(spriteHeader) + spr->hdr.number_frames * frame_size;

	// if we have a zero size, we have not yet allocated the resource
	if (index.size == 0)
		{
		// seek the starting location of the resource
		Seek(FileSize()); // ignore errors

		// and set up our data
		index.location = FilePos();
		index.size = size;
		}
	else
		{
		// check whether or not the size is okay
		if (size!=index.size) return fileError = resource_ResourceAlreadyUsed;

		// seek the starting location of the resource
		if ((fileError = Seek( index.location )) != resource_NoError) return fileError;
		}

//	DBG(370,"WRITE_RESOURCE:: WRITING %ld bytes TO THE RESOURCE",size);

//	if ((fileError = DOSFile::Write(buffer,size)) != resource_NoError) return fileError;

	// write the resource

	// write sprite header first
	if ((fileError = DOSFile::Write( (char*)&spr->hdr,sizeof(spriteHeader) )) != resource_NoError) return fileError;

	// then write each frame
	for (int i=0; i<spr->hdr.number_frames; i++)
		{
		if ((fileError = DOSFile::Write( (char*)spr->frame[i], frame_size )) != resource_NoError) return fileError;
		}

	// update the relevent index
	WriteResourceIndex( resource_num );

	// update the resource header
	ReadHeader();
	header.num_resources++;
	WriteHeader();

	return fileError;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		AddResource
//	DESC:		will write a resource to the end of a resource file
//
//	-------------------------------------------------------------------------
int SpriteResourceFile::AddResource(sprite* spr)
	{
	// write the resource into this resourcefile
	int fileError = WriteResource(NumResources(),spr);

	return fileError;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		LoadAllResources
//	DESC:		malloc each data item then load each data item of a resource
//
//	-------------------------------------------------------------------------
char** SpriteResourceFile::LoadAllResources()
	{
	// get number of resources to load
	int num_resources = NumResources();

	// malloc memory for list of pointers to each resource
	sprite** datalist = (sprite**) malloc(sizeof(sprite*) * num_resources);

	for (int i=0;i<num_resources;i++)
		{
		// get header for resource
		spriteHeader header = GetResourceHeader(i);

		// malloc memory for each resource
		datalist[i] = grCreateSprite(header.number_frames, header.width, header.height);

		// read resource into memory where datalist[i] is pointing
		fileError = ReadResource (i, datalist[i]);

		// on error
		if (fileError)
			grDrawString(0,400,"could not read resource error #%d",fileError);
		}

	// return pointer to whole list
	return (char**)datalist;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		ClearAllResources
//	DESC:		free each data item of a resource
//
//	-------------------------------------------------------------------------
void SpriteResourceFile::FreeAllResources(char** datalist)
	{
	int i;
	for (i=NumResources()-1;i>=0;i--)
		// free each resource
		grDestroySprite((sprite*)datalist[i]);

	// free whole list of pointers
	free(datalist);
	}
