//	-------------------------------------------------------------------------
//
//	TIMER.CPP
//
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for a timer ISR which
//					generates events into the EVENTQUEUE.
//
//					The ISR is pretty simple, simply incrementing a counter
//					in a number of different 'tickers'. Upon this counter
//					reaching a maximum value, the counter resets & an event
//					is generated into the EventQueue.
//
//					By doing this we can generate specific events
//					(such as cmdAnimate) at either a fixed rate or an
//					on-the-fly (user-specified) rate.
//
//					NOTE: * currently uses timer0 and updates the system clock
//						    due to inability to set rate for timer1 & timer2
//
//						  * When setting the system timer to 60Hz and then
//					  		exiting without resetting to 18.2Hz, XTREEGOLD
//							seems to run _much_ faster, BORLANDC _much_
//							slower... :-) 			.curious.
//
//	-------------------------------------------------------------------------

#include "timer.hpp"
#include "event.hpp"
#include "eventmgr.hpp"

#include "stdlib.h"
#include "conio.h"
#include "stdio.h"
#include "dos.h"

//
// NOTICE:: 	ONE of the following defines MUST be uncommented.
//				Whichever is selected is used as the base timer
//
#define		__USING_TIMER_0
//#define		__USING_TIMER_1
//#define		__USING_TIMER_2

// interrupt & port numbers to access the timer hardware

#define		TIMER_INT			0x1C	// the timer interrupt #
#define		INT_CONTROL			0x20   	// the interrupt controller

#define		TIMER_0				0x40	// counter 0 -- system timer/disk
#define		TIMER_1				0x41	// counter 1 -- DON'T USE!!!!! memory refresh
#define		TIMER_2				0x42	// counter 2 -- tape drive/speaker
#define		TIMER_CONTROL		0x43	// the 8254's control register

// defines for the timer settings

#define		TIMER0_CONTROL_WORD	0x3C	// timer 0, mode 2, binary, least/most
#define		TIMER1_CONTROL_WORD 0xAC	// timer 1, mode 2, binary, least/most
#define		TIMER2_CONTROL_WORD	0xBC	// timer 2, mode 2, binary, least/most

dostime_t	system_time;

// macros to return the LO & HI bytes of a word, as bytes

#define		LO_BYTE(n)			(n & 0x00FF)
#define		HI_BYTE(n)			((n>>8) & 0x00FF)

// --------------------------------------------------------------------------
//
// GLOBAL data used in the timer ticker functions
//
// --------------------------------------------------------------------------

// storage of previous timer ISR (Interrupt Service Routine)

void _interrupt (_far *Old_Timer_Isr)();

// storage of our timer event
TimerEvent	timer_EventInfo;

// internal flags

bool	timer_HANDLER_INSTALLED = false;	// a flag to detect if installed
bool	timer_TICKING_ON = false;			// a flag to toggle ticking on/off

// the structure used for each individual timer ticker

struct tickertype
	{
	uint	event;			// the event to place in the queue upon completion
	uint	max_ticks;		// the maximum number of ticks for the ticker
	uint	tick_counter;	// the tick counter.
	bool	ticking;		// whether or not we are in fact ticking!
	};

// storage area for the ticker bits & pieces

tickertype	ticker[TIMER_MAX_TICKERS];	// storage of the timer tickers
uint		num_tickers = 0;			// the number of tickers active
uint		active_ticker = 0;			// a counter used by the ISR.

// temporary debugging variables
//bool		tickers_activated[TIMER_MAX_TICKERS];
uint		ints_per_sec = 0;
uint 		int_iter = 0;


//	-------------------------------------------------------------------------
//
//	FUNC:		timer_ChangeInterruptRate()
//	DESC:		sets the hardware timer to generate interrupts at 'rate' Mhz
//				eg. timer_60Mhz
//
//	-------------------------------------------------------------------------
void timer_ChangeInterruptRate( uint rate )
	{
#ifdef __USING_TIMER_0
	// send the control word
	outp(TIMER_CONTROL, TIMER0_CONTROL_WORD);
	// write lo byte of the new rate to the counter register
	outp(TIMER_0, LO_BYTE(rate) );
	// write hi byte if the new rate to the counter register
	outp(TIMER_0, HI_BYTE(rate) );
#endif

#ifdef __USING_TIMER_1
	// send the control word
	outp(TIMER_CONTROL, TIMER1_CONTROL_WORD);
	// write lo byte of the new rate to the counter register
	outp(TIMER_1, LO_BYTE(rate) );
	// write hi byte if the new rate to the counter register
	outp(TIMER_1, HI_BYTE(rate) );
#endif

#ifdef __USING_TIMER_2
	// send the control word
	outp(TIMER_CONTROL, TIMER2_CONTROL_WORD);
	// write lo byte of the new rate to the counter register
	outp(TIMER_2, LO_BYTE(rate) );
	// write hi byte if the new rate to the counter register
	outp(TIMER_2, HI_BYTE(rate) );
#endif

	// (for now!)
	switch (rate)
		{
		case	timer_100Hz:	ints_per_sec = 100;		break;
		case	timer_90Hz:		ints_per_sec = 90;		break;
		case	timer_80Hz:		ints_per_sec = 80;		break;
		case	timer_70Hz:		ints_per_sec = 70;		break;
		case	timer_60Hz:		ints_per_sec = 60;		break;
		case	timer_50Hz:		ints_per_sec = 50;		break;
		case	timer_40Hz:		ints_per_sec = 40;		break;
		case	timer_30Hz:		ints_per_sec = 30;		break;
		case	timer_20Hz:		ints_per_sec = 20;		break;
		case	timer_18Hz:		ints_per_sec = 18;		break;
		}
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		timer_AddTicker()
//	DESC:		Adds a new timer ticker to the list of active tickers
//
//	PARAMS:		nticks 	- the number of ticks before the event is generated
//				event  	- the timer event which is generated
//
//	RETURNS:	uint	- a handle which may later be used to modify the
//						  state of the ticker. 0xFFFF == too many tickers
//
//	-------------------------------------------------------------------------
uint timer_AddTicker( uint nticks, uint event )
	{
	// if we haven't run out of timers
	if (num_tickers<TIMER_MAX_TICKERS)
		{
		// turn off timing (temporarily)
		timer_TICKING_ON = false;

		// set our new timer data
		ticker[num_tickers].tick_counter = 0;
		ticker[num_tickers].max_ticks = nticks;
		ticker[num_tickers].event = event;
		ticker[num_tickers].ticking = true;

		// increment our max. index
		num_tickers++;

		// and turn timing back on
		timer_TICKING_ON = true;

		// and return our timer number
		return num_tickers - 1;
		}
	else
		// return an 'invalid' flag
		return 0xFFFF;
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		timer_DelTicker()
//	DESC:		removes a specified timer ticker from the ticker list
//
//	PARAMS:		ticker	- the ticker handle (from AddTicker()) to delete
//
//	-------------------------------------------------------------------------
void timer_DelTicker( uint zticker )
	{
	// skip out if stupid info
	if (zticker>=num_tickers) return;

	// turn the timer off (temporarily)
	timer_TICKING_ON = false;

	// start at our first timer
	int pos = 0;
	int pos1 = 0;

	// skip till we find the appropriate timer
	while (pos<zticker)
		pos++;

	// and simply shuffle the rest of the timers!
	do
		{
		pos1 = pos + 1;	// for efficiency
		ticker[pos].tick_counter = ticker[pos1].tick_counter;
		ticker[pos].max_ticks = ticker[pos1].max_ticks;
		ticker[pos].event = ticker[pos1].event;

		pos++;
		}
	while (pos<num_tickers);

	// we now have one less timer
	if (num_tickers>0) num_tickers--;

	// turn timing back on
	timer_TICKING_ON = true;
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		timer_SetTicker()
//	DESC:		modifies a specified ticker, either speeding it up or slowing
//				it down.
//
//	PARAMS:		ticker	- the ticker to set (returned from timer_AddTicker())
//				nticks	- the number of ticks before the event is generated
//
//	-------------------------------------------------------------------------
void timer_SetTicker( uint zticker, uint nticks )
	{
	// skip out if stupid input
	if (zticker>=num_tickers) return;

	// and set the timer
	ticker[zticker].tick_counter = 0;
	ticker[zticker].max_ticks = nticks;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_ResetTicker()
//	DESC:		resets a specified ticker to 0
//
//	-------------------------------------------------------------------------
void timer_ResetTicker( uint zticker )
	{
	// skip out if stupid
	if (zticker>=num_tickers) return;

	// reset the timer's counter
	ticker[zticker].tick_counter = 0;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_StopTicker()
//	DESC:		stops a specified timer ticker from ticking
//
//	-------------------------------------------------------------------------
void timer_StopTicker( uint zticker )
	{
	// skip out if stupid
	if (zticker>=num_tickers) return;

	// stop the ticker
	ticker[zticker].ticking = false;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_StartsTicker()
//	DESC:		starts a specified timer ticker ticking
//
//	-------------------------------------------------------------------------
void timer_StartTicker( uint zticker )
	{
	// skip out if stupid
	if (zticker>=num_tickers) return;

	// start the ticker
	ticker[zticker].ticking = true;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_ISR()
//	DESC:		This is our timer interrupt function.
//				For simplicity, this function simply scans through all active
//				"timers" and increments their counters. If the ticks are all
//				completed an event is generated into the event queue!
//
//	-------------------------------------------------------------------------
void _interrupt _far timer_ISR()
	{
	if (timer_TICKING_ON)
		{
		// define our event
		timer_EventInfo.time = 0;

		// start with the first timer
		active_ticker = 0;

		// and update all tickers, generating events if required
		while (active_ticker<num_tickers)
			{
			// if the ticker is ticking
			if (ticker[active_ticker].ticking)
				// increment the timer's counter
				ticker[active_ticker].tick_counter++;

			// and if it has reached its threshold
			if (ticker[active_ticker].tick_counter>ticker[active_ticker].max_ticks)
					{
					// define our event
					timer_EventInfo.event = ticker[active_ticker].event;

					// if this event isn't in the queue
					if (!EVENTMGR->EventInQueue(ev_timer,&timer_EventInfo))
						// generate an event
						EVENTMGR->GenEvent(0,ev_timer,&timer_EventInfo);

					// and reset the timer
					ticker[active_ticker].tick_counter = 0;
					}

			// try the next timer
			active_ticker++;
			}
		}

	// here we manually keep track of system time!!!
	int_iter++;
	if (int_iter>ints_per_sec)
		{
		int_iter = 0;
		system_time.second++;
		if (system_time.second>59)
			{
			system_time.second = 0;
			system_time.minute++;
			if (system_time.minute>59)
				{
				system_time.minute = 0;
				system_time.hour++;
				if (system_time.hour>23)
					{
					system_time.hour = 0;
					// system_date.day +=;
					}
				}
			}
		}

	// chain to old ISR
	Old_Timer_Isr();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_InstallHandler()
//	DESC:		installs our timer ISR. Must call timer_RemoveHandler() upon
//				program completion
//
//	-------------------------------------------------------------------------
void timer_InstallHandler()
	{
	if (!timer_HANDLER_INSTALLED)
		{
		// save the old timer ISR
		Old_Timer_Isr = _dos_getvect( TIMER_INT );

		// and set our own timer ISR
		_dos_setvect( TIMER_INT, timer_ISR );

		// make sure its disabled on exit()
		atexit(timer_RemoveHandler);

		// set a flag to say we are installed
		timer_HANDLER_INSTALLED = true;

		// save the current system time
		_dos_gettime(&system_time);

		// by default, set the rate to 60Hz
		timer_ChangeInterruptRate(timer_60Hz);
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_RemoveHandler()
//	DESC:		removes our timer ISR, restoring the previous handler
//
//	-------------------------------------------------------------------------
void timer_RemoveHandler()
	{
	if (timer_HANDLER_INSTALLED)
		{
#ifdef	__USING_TIMER_0
		// reset the system time
		_dos_settime(&system_time);
#endif
		// set the timer back to 18.2 Hz
		timer_ChangeInterruptRate( timer_18Hz );

		// restore the old timer ISR
		_dos_setvect( TIMER_INT, Old_Timer_Isr );

		// and set an appropriate flag
		timer_HANDLER_INSTALLED = false;
		}
	}

/*
//#define cmdAnimate		5000
#define	cmdComputerAI	5001

#define evh_computer	6000

void main()
	{
	bool quit = false;

	clrscr();
	timer_InstallHandler();

	timer_AddTicker(50,cmdAnimate);	// animate every 50 ticks
	timer_AddTicker(100,cmdComputerAI);		// process computer AI every 100 ticks

	gotoxy(1,1);
	printf("Animation  ticker == %u\n",ticker[0].max_ticks);
	printf("ComputerAI ticker == %u\n",ticker[1].max_ticks);

//	timer_ChangeInterruptRate(timer_100Hz);

	uint iter = 0;
	uchar quitsec = system_time.second + 10;
	if (quitsec>59) quitsec -= 60;

	dostime_t	t;

	uint ticker0_cnt = 0;
	uint ticker1_cnt = 0;

	while (!quit)
		{
		printf(".");
//		iter++;
		quit = (system_time.second==quitsec);

		if (timer_EventInfo.event==cmdAnimate)  { ticker0_cnt++; timer_EventInfo.event = 0; }
		if (timer_EventInfo.event==cmdComputerAI) { ticker1_cnt++; timer_EventInfo.event = 0; }

//		gotoxy(1,5);
//		printf("DOS: %.2u:%.2u.%.2u",t.hour,t.minute,t.second);
//		gotoxy(1,6);
//		printf("PRG: %.2u:%.2u.%.2u",system_time.hour,system_time.minute,system_time.second);

//		gotoxy(28,1);
//		printf("cmdAnimate: %u,    cmdCompAI: %u   ",ticker[0].tick_counter,ticker[1].tick_counter);
//		gotoxy(28,2);
//		printf("%u   ",ticker[1].tick_counter);

//		if (any) delay(1000);

		if (kbhit())
			{
			switch (getch())
				{
				case	27:;	// ESC
				case	'q':;
				case	'Q':		quit = true; break;
				case	'1':		timer_ChangeInterruptRate(timer_18Hz);
									gotoxy(40,2); printf("Timer @ 18Hz ");
									break;
				case	'2':		timer_ChangeInterruptRate(timer_20Hz);
									gotoxy(40,2); printf("Timer @ 20Hz ");
									break;
				case	'3':		timer_ChangeInterruptRate(timer_30Hz);
									gotoxy(40,2); printf("Timer @ 30Hz ");
									break;
				case	'4':		timer_ChangeInterruptRate(timer_40Hz);
									gotoxy(40,2); printf("Timer @ 40Hz ");
									break;
				case	'5':		timer_ChangeInterruptRate(timer_50Hz);
									gotoxy(40,2); printf("Timer @ 50Hz ");
									break;
				case	'6':		timer_ChangeInterruptRate(timer_60Hz);
									gotoxy(40,2); printf("Timer @ 60Hz ");
									break;
				case	'7':		timer_ChangeInterruptRate(timer_70Hz);
									gotoxy(40,2); printf("Timer @ 70Hz ");
									break;
				case	'8':		timer_ChangeInterruptRate(timer_80Hz);
									gotoxy(40,2); printf("Timer @ 80Hz ");
									break;
				case	'9':		timer_ChangeInterruptRate(timer_90Hz);
									gotoxy(40,2); printf("Timer @ 90Hz ");
									break;
				case	'0':		timer_ChangeInterruptRate(timer_100Hz);
									gotoxy(40,2); printf("Timer @ 100Hz");
									break;
				case	'Z':		;
				case	'z':		timer_ChangeInterruptRate(timer_ZOOM);
									gotoxy(40,2); printf("Timer @ ZOOM");
									break;
				default		:;
				}
			}
		}

	printf("\nticker0 cnt == %d",ticker0_cnt);
	printf("\nticker1 cnt == %d",ticker1_cnt);

	timer_RemoveHandler();
	}

*/