//	-------------------------------------------------------------------------
//
//	TIMER.HPP
//
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains defines for a timer ISR which
//					generates events into the EVENTQUEUE.
//
//					The ISR is pretty simple, simply incrementing a counter
//					in a number of different 'tickers'. Upon this counter
//					reaching a maximum value, the counter resets & an event
//					is generated into the EventQueue.
//
//					By doing this we can generate specific events
//					(such as cmdAnimate) at either a fixed rate or an
//					on-the-fly (user-specified) rate.
//
//	-------------------------------------------------------------------------

#ifndef __timer_hpp
#define __timer_hpp

#include "stddefns.hpp"

// the maximum no. of tickers at once

#define		TIMER_MAX_TICKERS		16		// max. 16 tickers concurrent

// different constants for the timer (in hertz) used in timer_ChangeInterruptRate()

#define		timer_ZOOM			0x0100
#define		timer_100Hz			0x2E9B
#define		timer_90Hz			0x33C9
#define		timer_80Hz			0x3A42
#define		timer_70Hz			0x4295
#define 	timer_60Hz			0x4DAE
#define		timer_50Hz			0x5D37
#define		timer_40Hz			0x7486
#define		timer_30Hz			0x965C
#define		timer_20Hz			0xE90B
#define		timer_18Hz			0xFFFF	// the standard timer setting 18.2Hz

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_ChangeInterruptRate()
//	DESC:		sets the hardware timer to generate interrupts at 'rate' Mhz
//				eg. timer_60Mhz
//
//	-------------------------------------------------------------------------
void timer_ChangeInterruptRate( uint rate );

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_AddTicker()
//	DESC:		Adds a new timer ticker to the list of active tickers
//
//	PARAMS:		nticks 	- the number of ticks before the event is generated
//				event  	- the timer event which is generated
//
//	RETURNS:	uint	- a handle which may later be used to modify the
//						  state of the ticker. 0xFFFF == too many tickers
//
//	-------------------------------------------------------------------------
uint timer_AddTicker( uint nticks, uint event );

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_DelTicker()
//	DESC:		removes a specified timer ticker from the ticker list
//
//	PARAMS:		ticker	- the ticker handle (from AddTicker()) to delete
//
//	-------------------------------------------------------------------------
void timer_DelTicker( uint zticker );

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_SetTicker()
//	DESC:		modifies a specified ticker, either speeding it up or slowing
//				it down.
//
//	PARAMS:		ticker	- the ticker to set (returned from timer_AddTicker())
//				nticks	- the number of ticks before the event is generated
//
//	-------------------------------------------------------------------------
void timer_SetTicker( uint zticker, uint nticks );

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_ResetTicker()
//	DESC:		resets a specified ticker to 0
//
//	-------------------------------------------------------------------------
void timer_ResetTicker( uint zticker );

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_StopTicker()
//	DESC:		stops a specified timer ticker from ticking
//
//	-------------------------------------------------------------------------
void timer_StopTicker( uint zticker );

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_StartsTicker()
//	DESC:		starts a specified timer ticker ticking
//
//	-------------------------------------------------------------------------
void timer_StartTicker( uint zticker );

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_InstallHandler()
//	DESC:		installs our timer ISR. Must call timer_RemoveHandler() upon
//				program completion
//
//	-------------------------------------------------------------------------
void timer_InstallHandler();

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_RemoveHandler()
//	DESC:		removes our timer ISR, restoring the previous handler
//
//	-------------------------------------------------------------------------
void timer_RemoveHandler();


#endif